import inspect
from types import SimpleNamespace

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd
import pytest
import torch
from sklearn.linear_model import LinearRegression

from causalexplain.explainability import hierarchies as hmod
from causalexplain.explainability import perm_importance as pimod
from causalexplain.explainability import regression_quality as rqmod
from causalexplain.explainability import shapley as smod


class DummyProgBar:
    def start_subtask(self, name, total):
        return self

    def update_subtask(self, name, value):
        return self

    def remove(self, name):
        return self


@pytest.fixture(autouse=True)
def patch_progbar(monkeypatch):
    monkeypatch.setattr(hmod, "ProgBar", DummyProgBar, raising=False)
    monkeypatch.setattr(pimod, "ProgBar", DummyProgBar, raising=False)
    monkeypatch.setattr(smod, "ProgBar", DummyProgBar, raising=False)


def test_hierarchies_fit_clusters_and_dissimilarity():
    data = pd.DataFrame(
        {
            "a": [1.0, 2.0, 3.0, 4.0],
            "b": [1.1, 2.1, 2.9, 4.1],
            "c": [4.0, 3.0, 2.0, 1.0],
        }
    )
    h = hmod.Hierarchies(method="spearman", correlation_th=0.5)
    h.fit(data)

    # Exercise clustering branches
    clustered_corr, sorted_cols = h._cluster_features("spearman", threshold=0.5)
    assert set(sorted_cols) == set(data.columns)
    assert clustered_corr.shape == (3, 3)

    dis = h.hierarchical_dissimilarities()
    assert set(dis.columns) == set(data.columns)
    assert np.all(dis.values >= 0)


def test_hierarchies_helpers_and_plot(monkeypatch):
    data = pd.DataFrame({"a": [1, 2, 1.5], "b": [2, 1, 3]})
    corr = hmod.Hierarchies.compute_correlation_matrix(data, method="spearman")
    correlated = hmod.Hierarchies.compute_correlated_features(
        corr, 0.1, ["a", "b"], verbose=True
    )
    assert isinstance(correlated, dict)

    linkage_mat = np.array([[0, 1, 0.5, 2]])
    clusters = hmod.Hierarchies()._clusters_from_linkage(linkage_mat, ["a", "b"])
    assert hmod.Hierarchies._get_cluster(clusters, "a") is not None
    assert not hmod.Hierarchies._is_cluster("a")
    assert hmod.Hierarchies._contains_a_cluster(clusters, "K2") is False
    assert hmod.Hierarchies._get_cluster_element(clusters, "K2") is None
    assert hmod.Hierarchies._in_cluster(clusters["K2"], "a")
    assert hmod.Hierarchies._in_same_cluster(clusters, "a", "b") == "K2"
    assert hmod.Hierarchies()._are_connected(clusters, "a", "b") == pytest.approx(0.5)

    monkeypatch.setattr(hmod.plt, "show", lambda *args, **kwargs: None)
    Z = hmod.plot_dendogram_correlations(corr, ["a", "b"], figsize=(2, 2))
    assert Z is not None


def test_hierarchies_expand_clusters_and_connect_verbose(monkeypatch):
    h = hmod.Hierarchies()
    h.data = pd.DataFrame({"a": [1, 2], "b": [2, 1]})
    h.linkage_mat = np.array([[0, 1, 0.2, 2]])
    pi = SimpleNamespace(
        feature_importances_={
            "a": {"b": 0.1},
            "b": {"a": 0.2},
        },
        regression_importances_=[0.5, 0.6],
    )
    ground_truth = nx.DiGraph([("a", "b")])
    h.expand_clusters_perm_importance(pi, ground_truth)

    G = nx.DiGraph()
    G.add_edge("a", "b")
    linkage_mat = np.array([[0, 1, 0.1, 2]])
    conn = hmod.connect_hierarchies(G, linkage_mat, ["a", "b"], verbose=True)
    assert isinstance(conn, nx.DiGraph)


def test_hierarchies_correlated_features_verbose():
    corr = pd.DataFrame([[1.0, 0.5], [0.5, 1.0]], columns=["a", "b"], index=["a", "b"])
    correlated = hmod.Hierarchies.compute_correlated_features(corr, 0.4, ["a", "b"], verbose=True)
    assert correlated["a"] == ["b"]


def test_hierarchies_correlated_features_no_threshold():
    corr = pd.DataFrame(np.eye(2), columns=["a", "b"], index=["a", "b"])
    correlated = hmod.Hierarchies.compute_correlated_features(corr, None, ["a", "b"])
    assert correlated == {}


def test_hierarchies_nested_cluster_connection():
    clusters = {
        "K2": [("a", "K3"), 0.4],
        "K3": [("b", "c"), 0.7],
    }
    h = hmod.Hierarchies()
    assert h._are_connected(clusters, "a", "b") == 0.4


def test_connect_isolated_nodes_skips_large_clusters():
    G = nx.DiGraph()
    linkage_mat = np.array([[0, 1, 0.1, 3]])
    result = hmod.connect_isolated_nodes(G, linkage_mat, ["a", "b"], verbose=True)
    assert isinstance(result, nx.DiGraph)


def test_regression_quality_predict_and_verbose(capsys):
    assert isinstance(rqmod.RegQuality(), rqmod.RegQuality)
    scores = [0.1, 0.2, 5.0, 6.0]
    outliers = rqmod.RegQuality.predict(scores, gamma_shape=1, gamma_scale=1, threshold=0.9)
    assert outliers == set()

    _ = rqmod.RegQuality._mad_criteria(scores, verbose=True)
    verbose_out = capsys.readouterr().out
    assert "Median" in verbose_out

    gamma_indices = rqmod.RegQuality._gamma_criteria(scores, threshold=0.2, verbose=True)
    assert gamma_indices == {2, 3}


def _build_simple_models(df):
    # Simple container with regressors and predict
    class SimpleModels:
        def __init__(self, data):
            self.regressor = {}
            for col in data.columns:
                X = data.drop(columns=[col]).values
                y = data[col].values
                # Lightweight sklearn-style regressor
                lr = LinearRegression().fit(X, y)
                self.regressor[col] = lr

        def predict(self, X):
            return X.values

    return SimpleModels(df)


def test_permutation_importance_sklearn_flow(monkeypatch):
    df = pd.DataFrame(
        {
            "x": [1.0, 2.0, 3.0, 4.0],
            "y": [1.1, 2.1, 2.9, 4.1],
            "z": [0.5, 0.4, 0.6, 0.7],
        }
    )
    models = _build_simple_models(df)
    original_add_zeroes = pimod.PermutationImportance._add_zeroes

    def fake_add_zeroes(self, target, correlated_features):
        # track that we hit correlation branch
        self._zeroed = (target, tuple(correlated_features))
        return original_add_zeroes(self, target, correlated_features)

    # Avoid building DAGs via utils; only return edges passed in.
    def fake_digraph(X, feature_names, models_arg, connections, root_causes, reciprocity=True, anm_iterations=10, verbose=False):
        g = nx.DiGraph()
        for src, targets in connections.items():
            for tgt in targets:
                g.add_edge(tgt, src)
        return g

    monkeypatch.setattr(pimod.utils, "digraph_from_connected_features", fake_digraph)
    monkeypatch.setattr(pimod.PermutationImportance, "_add_zeroes", fake_add_zeroes, raising=False)

    pi = pimod.PermutationImportance(
        models=models,
        correlation_th=0.8,
        n_repeats=1,
        mean_pi_percentile=0.5,
        prog_bar=False,
        verbose=False,
    )
    pi.fit(df)
    graph = pi.predict(df)
    assert isinstance(graph, nx.DiGraph)
    assert pi.mean_pi_threshold >= 0.0
    assert hasattr(pi, "_zeroed")

    # Plot uses computed metrics and should not crash
    fig, ax = plt.subplots(1, 1, figsize=(3, 2))
    returned = pi._plot_perm_imp("x", ax=ax)
    assert returned is not None


def test_shuffle_tensor_column_and_loss_computation():
    tensor = torch.tensor([[1.0, 2.0], [3.0, 4.0]])
    pi = pimod.PermutationImportance(models=SimpleNamespace(regressor={"a": None}))
    shuffled = pi._shuffle_2Dtensor_column(tensor, 1)
    assert shuffled.shape == tensor.shape
    # Dummy model/dataloader to exercise loss computation
    class DummyModel(torch.nn.Module):
        def __init__(self):
            super().__init__()
            self.loss_fn = torch.nn.MSELoss()

        def forward(self, x):
            return x.sum(dim=1, keepdim=True)

    dataset = [(torch.tensor([1.0, 2.0]), torch.tensor([3.0]))]
    loader = torch.utils.data.DataLoader(dataset, batch_size=1)
    model = DummyModel()
    avg, std, losses = pi._compute_loss_shuffling_column(model, loader, shuffle_col=0)
    assert losses.size >= 1
    assert avg == pytest.approx(losses.mean())


def test_compute_perm_imp_repeats(monkeypatch):
    class DummyReg:
        def __init__(self):
            self.columns = ["f1", "f2", "f3"]
            X = torch.tensor([[1.0, 2.0, 3.0]])
            y = torch.tensor([[6.0]])
            self.train_loader = torch.utils.data.DataLoader(list(zip(X, y)), batch_size=1)
            self.model = DummyModel()

    class DummyModel(torch.nn.Module):
        def __init__(self):
            super().__init__()
            self.loss_fn = torch.nn.MSELoss()

        def forward(self, x):
            return x.sum(dim=1, keepdim=True)

    reg = DummyReg()
    pi = pimod.PermutationImportance(models=SimpleNamespace(regressor={"t": reg}), n_repeats=2, prog_bar=False, verbose=False)
    pi.base_loss = {"t": 0.0}
    mean, std = pi._compute_perm_imp("t", reg, reg.model, num_vars=3)
    assert mean.shape[0] == 2


def test_permutation_importance_corr_branch_plot(monkeypatch):
    df = pd.DataFrame({"x": [1.0, 2.0, 3.0], "y": [1.1, 2.1, 3.1], "z": [0.5, 0.6, 0.7]})
    models = _build_simple_models(df)
    monkeypatch.setitem(pimod.plt.rcParams, "text.usetex", False)
    import matplotlib.texmanager as texmanager
    monkeypatch.setattr(texmanager.TexManager, "_run_checked_subprocess", lambda *a, **k: None)
    monkeypatch.setattr(pimod, "subplots", lambda fn, *args, **kwargs: "stubbed")
    pi = pimod.PermutationImportance(
        models=models, correlation_th=0.5, n_repeats=1, prog_bar=False, verbose=True
    )
    pi._fit_sklearn(df)
    # Ensure correlation branches set correlated_features
    assert any(len(v) >= 0 for v in pi.correlated_features.values())
    pi.connections = {name: ["dummy"] for name in pi.feature_names}
    assert pi.plot(figsize=(2, 2)) == "stubbed"


def test_permutation_importance_obtain_corr_info():
    df = pd.DataFrame({"x": [1.0, 2.0, 3.0], "y": [1.1, 2.1, 3.1]})
    models = _build_simple_models(df)
    pi = pimod.PermutationImportance(models=models, correlation_th=0.5, prog_bar=False, verbose=False)
    pi._obtain_correlation_info(df)
    assert pi.corr_matrix is not None


def test_permutation_importance_no_corr_branch():
    df = pd.DataFrame({"x": [1.0, 2.0], "y": [2.0, 3.0]})
    models = _build_simple_models(df)
    pi = pimod.PermutationImportance(models=models, correlation_th=None, prog_bar=False, verbose=False)
    pi.fit(df)


def test_permutation_importance_add_zeroes_branch(monkeypatch):
    pi = pimod.PermutationImportance(models=SimpleNamespace(regressor={"x": None, "y": None}), prog_bar=False, verbose=False)
    pi.feature_names = ["x", "y"]
    pi.pi = {"x": {"importances_mean": np.array([0.1]), "importances_std": np.array([0.01])}}
    pi._add_zeroes("x", ["y"])
    assert pi.pi["x"]["importances_mean"].shape[0] == 2


def test_permutation_importance_pytorch_path(monkeypatch):
    # Dummy torch-style regressor/model
    class TorchReg:
        def __init__(self):
            self.model = DummyModel()
            self.columns = ["f1", "f2"]
            X = torch.tensor([[1.0, 2.0], [3.0, 4.0]])
            y = torch.tensor([[3.0], [7.0]])
            self.train_loader = torch.utils.data.DataLoader(list(zip(X, y)), batch_size=1)

    class DummyModel(torch.nn.Module):
        def __init__(self):
            super().__init__()
            self.loss_fn = torch.nn.MSELoss()

        def forward(self, x):
            return x.sum(dim=1, keepdim=True)

    reg = TorchReg()
    models = SimpleNamespace(regressor={"a": reg, "b": reg})

    monkeypatch.setattr(
        pimod.utils,
        "valid_candidates_from_prior",
        lambda feats, tgt, prior: [f for f in feats if f != tgt],
    )
    monkeypatch.setattr(pimod.utils, "digraph_from_connected_features", lambda X, *args, **kwargs: nx.DiGraph())
    monkeypatch.setattr(pimod.utils, "break_cycles_if_present", lambda g, *args, **kwargs: g)

    pi = pimod.PermutationImportance(
        models=models,
        n_repeats=1,
        prog_bar=False,
        verbose=True,
    )
    pi._fit_pytorch()
    pi.mean_pi_threshold = 0.0
    pi.prior = []
    g = pi._predict_pytorch(pd.DataFrame({"a": [1, 2], "b": [3, 4]}), root_causes=None)
    assert isinstance(g, nx.DiGraph)


def test_permutation_importance_fit_dispatch(monkeypatch):
    class TorchReg:
        def __init__(self):
            self.model = DummyModel()
            self.columns = ["x", "y"]
            X = torch.tensor([[1.0, 2.0]])
            y = torch.tensor([[3.0]])
            self.train_loader = torch.utils.data.DataLoader(list(zip(X, y)), batch_size=1)

    class DummyModel(torch.nn.Module):
        def __init__(self):
            super().__init__()
            self.loss_fn = torch.nn.MSELoss()

        def forward(self, x):
            return x.sum(dim=1, keepdim=True)

    monkeypatch.setattr(pimod, "MLPModel", TorchReg)
    models = SimpleNamespace(regressor={"x": TorchReg()})
    pi = pimod.PermutationImportance(models=models, prog_bar=False, verbose=False, n_repeats=1)
    pi.fit(pd.DataFrame({"x": [1.0], "y": [2.0]}))


def test_shap_estimator_fit_predict_and_adjust(monkeypatch):
    df = pd.DataFrame(
        {
            "a": [1.0, 2.0, 3.0, 4.0],
            "b": [1.5, 2.5, 3.5, 4.5],
            "c": [0.5, 1.0, 1.5, 2.0],
        }
    )

    class FakeModels:
        def __init__(self, columns):
            self.regressor = {col: SimpleNamespace(predict=lambda X: np.ones(len(X))) for col in columns}

        def predict(self, X):
            return X.values

    models = FakeModels(df.columns)

    def fake_run_explainer(self, target_name, model, X_train, X_test):
        # Return a small, deterministic SHAP matrix
        num_features = X_train.shape[1]
        return np.tile(np.arange(1, num_features + 1), (len(X_test), 1))

    monkeypatch.setattr(smod.ShapEstimator, "_run_selected_shap_explainer", fake_run_explainer)
    monkeypatch.setattr(smod.utils, "valid_candidates_from_prior", lambda feats, t, p: [f for f in feats if f != t])
    monkeypatch.setattr(
        smod.utils,
        "digraph_from_connected_features",
        lambda X, feature_names, models_arg, connections, root_causes, prior=None, reciprocity=False, anm_iterations=10, verbose=False: nx.DiGraph(
            [(parent, target) for target, parents in connections.items() for parent in parents]
        ),
    )
    monkeypatch.setattr(smod.utils, "break_cycles_if_present", lambda g, *args, **kwargs: g)

    sh = smod.ShapEstimator(models=models, mean_shap_percentile=0.5, prog_bar=False, verbose=False)
    sh.fit(df)
    dag = sh.predict(df)
    assert isinstance(dag, nx.DiGraph)
    assert sh.mean_shap_threshold >= 0.0
    assert sh.discrepancies is not None

    # Adjust edges based on discrepancies to exercise reversal logic
    graph = nx.DiGraph()
    graph.add_edges_from([("a", "b"), ("b", "a")])
    adjusted = sh._adjust_edges_from_shap_discrepancies(graph, increase_tolerance=0.1, sd_upper=1.0)
    assert isinstance(adjusted, nx.DiGraph)

    # Plot summary uses computed fields
    fig = sh._plot_shap_summary("a", ax=None, figsize=(3, 2))
    assert fig is not None


def test_shap_helpers_vector_and_error_contribution():
    sh = smod.ShapEstimator(models=None, prog_bar=False, verbose=False)
    discrepancies = pd.DataFrame({"a": [0.1, 0.2], "b": [0.2, 0.1]}, index=["a", "b"])
    vec = sh._input_vector(discrepancies, "a", "b", target_mean=0.15)
    assert vec.shape == (7,)

    predictions = [np.array([1.0, 2.0]), np.array([3.0, 4.0])]
    reshaped = sh._adjust_predictions_shape(predictions, (2, 2))
    assert reshaped.shape == (2, 2)

    shap_values = pd.DataFrame({"a": [0.1, 0.2], "b": [0.0, 0.1]})
    y_true = pd.Series([1.0, 2.0])
    y_pred = pd.Series([1.0, 2.5])
    contribution = sh._individual_error_contribution(shap_values, y_true, y_pred)
    assert "a" in contribution.index

    # Debug message should be silent when verbose=False
    sh._debugmsg("Ignored edge", "a", 0.2, "b", 0.0, [0.1, 0.2, 0.1, 0, 0, 0, 0], 0.5, [])


def test_shap_get_method_caller_name_handles_unknown(monkeypatch):
    sh = smod.ShapEstimator(models=None, prog_bar=False, verbose=False)
    # Force inspect.currentframe to raise
    monkeypatch.setattr(inspect, "currentframe", lambda: (_ for _ in ()).throw(RuntimeError("fail")))
    assert sh._get_method_caller_name() == "unknown"


def test_shap_discrepancy_value_error_paths(monkeypatch):
    sh = smod.ShapEstimator(models=None, prog_bar=False, verbose=False)
    monkeypatch.setattr(smod.sms, "het_breuschpagan", lambda *args, **kwargs: (_ for _ in ()).throw(ValueError("fail")))
    disc = sh._compute_discrepancy(np.array([1, 2]), np.array([1, 2]), np.array([[1], [2]]), "t", "p")
    assert isinstance(disc, smod.ShapDiscrepancy)


def test_shap_fit_target_uses_model_attr():
    class WithModel:
        def __init__(self):
            self.model = SimpleNamespace(cpu=lambda: self, predict=lambda X: np.ones(len(X)), cuda=lambda: self)
            self.predict = lambda X: np.ones(len(X))

    models = SimpleNamespace(regressor={"a": WithModel(), "b": SimpleNamespace(predict=lambda X: np.ones(len(X)))})
    res = smod.ShapEstimator._shap_fit_target_variable(
        "a",
        models,
        pd.DataFrame({"a": [1, 2], "b": [3, 4]}),
        pd.DataFrame({"a": [5, 6], "b": [7, 8]}),
        ["a", "b"],
        on_gpu=True,
        verbose=True,
        run_selected_shap_explainer_func=lambda *args, **kwargs: np.array([[1.0], [2.0]]),
    )
    assert res[0] == "a"


def test_shap_explainer_variants(monkeypatch):
    X_train = np.array([[1.0, 2.0], [3.0, 4.0]])
    X_test = np.array([[5.0, 6.0]])

    class DummyModel:
        def predict(self, X):
            return np.sum(X, axis=1)

        def to(self, device):
            return self

    # Kernel branch
    sh = smod.ShapEstimator(models=None, prog_bar=False, verbose=False)
    sh.explainer = "kernel"
    sh.shap_explainer = {}
    monkeypatch.setattr(
        smod.shap,
        "KernelExplainer",
        lambda predict, data: SimpleNamespace(shap_values=lambda x: [np.ones((len(x), data.shape[1]))]),
    )
    kernel_vals = sh._run_selected_shap_explainer("t", DummyModel(), X_train, X_test)
    assert kernel_vals.shape == (1, 2)

    # Gradient branch
    sh.explainer = "gradient"
    monkeypatch.setattr(
        smod.shap,
        "GradientExplainer",
        lambda model, data: lambda x: SimpleNamespace(values=np.full((len(x), data.shape[1]), 2.0)),
    )
    grad_vals = sh._run_selected_shap_explainer("t", DummyModel(), X_train, X_test)
    assert grad_vals.shape == (1, 2)

    # Explainer branch
    sh.explainer = "explainer"
    monkeypatch.setattr(
        smod.shap,
        "Explainer",
        lambda predict, data: lambda x, silent=True: SimpleNamespace(values=np.full((len(x), data.shape[1]), 3.0)),
    )
    exp_vals = sh._run_selected_shap_explainer("t", DummyModel(), X_train, X_test)
    assert exp_vals.shape == (1, 2)

    sh.explainer = "unknown"
    with pytest.raises(ValueError):
        sh._run_selected_shap_explainer("t", DummyModel(), X_train, X_test)


def test_shap_threshold_and_zero_injection():
    sh = smod.ShapEstimator(models=None, prog_bar=False, verbose=False)
    sh.mean_shap_percentile = None
    sh.all_mean_shap_values = np.array([1.0, 2.0])
    sh._compute_scaled_shap_threshold()
    assert sh.mean_shap_threshold == 0.0

    sh.feature_names = ["a", "b", "c"]
    sh.all_mean_shap_values = [np.array([0.2, 0.3])]
    sh._add_zeroes("a", ["b"])
    assert len(sh.all_mean_shap_values[-1]) == 3


def test_shap_verbose_debug_and_tolerance(monkeypatch):
    sh = smod.ShapEstimator(models=None, prog_bar=False, verbose=True)
    # Force verbose paths inside helpers
    discrepancies = pd.DataFrame([[0.1, 0.2], [0.05, 0.1]], columns=["a", "b"], index=["a", "b"])
    assert not sh._increase_upper_tolerance(discrepancies)
    big_disc = pd.DataFrame([[10.0, 0.0], [0.0, 0.0]], columns=["a", "b"], index=["a", "b"])
    assert sh._increase_upper_tolerance(big_disc)

    cycles = [["x", "y", "x"]]
    assert sh._nodes_in_cycles(cycles, "x", "y")

    sh.feature_names = ["a", "b"]
    sh.discrepancies = discrepancies
    graph = nx.DiGraph([("a", "b")])
    sh._debugmsg("Ignored edge", "a", 0.2, "b", 0.0, [0.1, 0.2, 0.1, 0, 0, 0, 0], 0.5, [])
    sh._debugmsg("(*) Reversed edge", "a", 0.2, "b", 0.0, [0.1, 0.2, 0.1, 0, 0, 0, 0], 0.5, cycles)
    sh._adjust_edges_from_shap_discrepancies(graph, increase_tolerance=0.0, sd_upper=0.5)


def test_shap_fit_parallel_branch(monkeypatch):
    df = pd.DataFrame({"a": [1.0, 2.0], "b": [2.0, 3.0]})
    models = _build_simple_models(df)

    # Force use of multiprocessing branch while avoiding real pools
    class FakePool:
        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc, tb):
            return False

        def imap_unordered(self, func, iterable):
            for target in iterable:
                yield (target, np.ones((2, 1)), np.array([0]), np.array([0.2, 0.3]))

        def close(self):
            return None

        def join(self):
            return None

    class FakeCtx:
        def Pool(self, processes):
            return FakePool()

    monkeypatch.setattr(smod, "get_context", lambda *args, **kwargs: FakeCtx())
    monkeypatch.setattr(
        smod.ShapEstimator,
        "_shap_fit_target_variable",
        lambda *args, **kwargs: ("a", np.ones((2, 1)), np.array([0]), np.array([0.1, 0.2])),
    )

    sh = smod.ShapEstimator(models=models, mean_shap_percentile=0.5, prog_bar=False, verbose=False, parallel_jobs=1)
    sh.fit(df)
    assert sh.is_fitted_


def test_shap_fit_progress_bar(monkeypatch):
    df = pd.DataFrame({"a": [1.0, 2.0, 3.0], "b": [2.0, 3.0, 4.0]})
    models = _build_simple_models(df)
    sh = smod.ShapEstimator(models=models, mean_shap_percentile=0.5, prog_bar=True, verbose=False, parallel_jobs=0)
    sh.fit(df)
    assert sh.mean_shap_threshold >= 0


def test_shap_custom_main_and_sachs_main(monkeypatch):
    # Avoid filesystem access by mocking dependencies
    monkeypatch.setattr(smod.utils, "graph_from_dot_file", lambda *args, **kwargs: nx.DiGraph())
    monkeypatch.setattr(pd, "read_csv", lambda *args, **kwargs: pd.DataFrame({"a": [1, 2], "b": [2, 3]}))
    monkeypatch.setattr(
        smod, "StandardScaler", lambda *args, **kwargs: SimpleNamespace(fit_transform=lambda x: x)
    )

    class DummyModels(SimpleNamespace):
        def __init__(self):
            super().__init__()
            self.regressor = {"a": SimpleNamespace(predict=lambda X: np.ones(len(X)))}
            self.root_causes = []

    dummy_rex = SimpleNamespace(models=DummyModels(), root_causes=[], shaps=None, is_fitted_=True)
    monkeypatch.setattr(
        smod.utils, "load_experiment", lambda *args, **kwargs: dummy_rex
    )
    monkeypatch.setattr(smod, "ShapEstimator", lambda *args, **kwargs: SimpleNamespace(fit=lambda *a, **k: None, predict=lambda *a, **k: None))

    smod.custom_main("dummy", path="/tmp/", output_path="/tmp/", scale=True)
    smod.sachs_main()


def test_shap_predict_progress_bar(monkeypatch):
    df = pd.DataFrame({"a": [1.0, 2.0], "b": [2.0, 3.0]})
    models = _build_simple_models(df)

    def fake_run_explainer(self, target_name, model, X_train, X_test):
        # Shape with trailing singleton dimension to trigger reshape
        return np.ones((len(X_test), X_train.shape[1], 1))

    monkeypatch.setattr(smod.ShapEstimator, "_run_selected_shap_explainer", fake_run_explainer)
    monkeypatch.setattr(smod.utils, "valid_candidates_from_prior", lambda feats, t, p: [f for f in feats if f != t])
    monkeypatch.setattr(
        smod.utils,
        "digraph_from_connected_features",
        lambda X, feature_names, models_arg, connections, root_causes, prior=None, reciprocity=False, anm_iterations=10, verbose=False: nx.DiGraph(),
    )
    monkeypatch.setattr(smod.utils, "break_cycles_if_present", lambda g, *args, **kwargs: g)

    sh = smod.ShapEstimator(models=models, mean_shap_percentile=0.5, prog_bar=True, verbose=False)
    sh.fit(df)
    sh.predict(df)



import numpy as np
import pandas as pd
import pytest

from causalexplain.estimators.pc import ci_tests


def test_power_divergence_invalid_z():
    df = pd.DataFrame({"X": [0, 1], "Y": [1, 0]})
    with pytest.raises(Exception):
        ci_tests.power_divergence("X", "Y", Z=None, data=df)


def test_power_divergence_boolean_and_tuple():
    df = pd.DataFrame({"X": [0, 0, 1, 1], "Y": [0, 0, 1, 1]})
    chi, p_value, dof = ci_tests.power_divergence("X", "Y", Z=[], data=df, boolean=False)
    assert chi >= 0
    assert 0 <= p_value <= 1
    assert dof == 1

    independent = ci_tests.power_divergence("X", "Y", Z=[], data=df, boolean=True, significance_level=0.0)
    assert isinstance(independent, (bool, np.bool_))


def test_power_divergence_with_conditionals():
    df = pd.DataFrame(
        {
            "X": [0, 0, 1, 1, 0, 1],
            "Y": [0, 1, 0, 1, 0, 1],
            "Z": [0, 0, 0, 0, 1, 1],
        }
    )
    result = ci_tests.power_divergence("X", "Y", Z=["Z"], data=df, boolean=True, significance_level=0.5)
    assert isinstance(result, (bool, np.bool_))


def test_pearsonr_validation_errors():
    df = pd.DataFrame({"X": [0], "Y": [1]})
    with pytest.raises(ValueError):
        ci_tests.pearsonr("X", "Y", Z=None, data=df)
    with pytest.raises(ValueError):
        ci_tests.pearsonr("X", "Y", Z=[], data=[[1, 2]])


def test_pearsonr_boolean_mode_and_residuals():
    rng = np.random.RandomState(0)
    df = pd.DataFrame({"X": rng.randn(50), "Y": rng.randn(50) * 0.5 + 1, "Z": rng.randn(50)})

    result_bool = ci_tests.pearsonr("X", "Y", Z=[], data=df, boolean=True, significance_level=0.01)
    assert isinstance(result_bool, bool)

    coef, p_val = ci_tests.pearsonr("X", "Y", Z=["Z"], data=df, boolean=False)
    assert abs(coef) < 0.3
    assert 0 <= p_val <= 1



from typing import List

import numpy as np
import pandas as pd
from pandas import DataFrame
from pygam import LinearGAM
from sklearn.gaussian_process import GaussianProcessRegressor as gpr

from causalexplain.independence.hsic import HSIC


def _fit_GPR(x, y, x_test, y_test):
    model = gpr(normalize_y=False)
    model.fit(x, y)
    prediction = model.predict(x_test).reshape(-1, 1)
    residuals = y_test - prediction
    return residuals


def _fit_GAM(x, y, x_test, y_test):
    model = LinearGAM()
    model.gridsearch(x, y, progress=False)
    pred = model.predict(x_test)
    residuals = np.array([x_test[i] - pred[i] for i in range(x_test.shape[0])])
    return residuals


def fit_and_get_residuals(
    X: np.ndarray,
    Y: np.ndarray,
    X_test: np.ndarray = None,
    Y_test: np.ndarray = None,
    method="gpr"
):
    """
    Fit a model y ~ f(X), where X is an independent variable and Y is a
    dependent one. The model is passed as argument, together with the
    training and test sets.

    Args:
        X: (np.ndarray) The feature to be used as input to predict Y_train
        Y: (np.ndarray) The feature to be predicted
        X_test: (np.ndarray) The feature to be used as input to predict Y_test
        Y_test: (np.ndarray) The feature to be predicted
        method: (str) Either "gpr" or "gam"

    Returns:
        The method returns the residuals and the RMS error.
    """
    # Fix dimensions if hasn't been done already.
    if np.ndim(X) == 1:
        X = X.reshape(-1, 1)
    if np.ndim(Y) == 1:
        Y = Y.reshape(-1, 1)
    if np.ndim(X_test) == 1:
        X_test = X_test.reshape(-1, 1)
    if np.ndim(Y_test) == 1:
        Y_test = Y_test.reshape(-1, 1)

    noise = np.random.normal(0, .1, X.shape[0]).reshape(-1, 1)
    Y = Y + noise

    if X_test is None or Y_test is None:
        X_test = X
        Y_test = Y

    if method == "gpr":
        # residuals, XX, YY = _fit_GPR(X, Y, X_test, Y_test)
        residuals = _fit_GPR(X, Y, X_test, Y_test)
    elif method == "gam":
        # residuals, XX, YY = _fit_GAM(X, Y, X_test, Y_test)
        residuals = _fit_GAM(X, Y, X_test, Y_test)
    else:
        raise ValueError(f"Invalid method: {method}")

    return residuals


def run_feature_selection(X: DataFrame, y: str) -> List:
    """
    Extracts 'y' from the list of features of "X" and call the prediction
    method passed to asses the predictive influence of each variable in X
    to obtain "y".

    Args:
         X: Dataframe with ALL continous variables
         y: the name of the variable in X to be used as target.
         predict_method: the method used to predict "y" from "X".
            "hsiclasso" or "block_hsic_lasso"

    Return:
        List: with the predictive score for each variable.
    """
    feature_names = list(X.columns.values)
    feature_names.remove(y)
    df_target = pd.DataFrame(X[y], columns=[y])
    df_features = X[feature_names]

    y = np.transpose(df_target.values)
    X = np.transpose(df_features.values)
    hsic = HSIC().fit(X, y)
    return hsic.stat



"""
Module for the edge orientation algorithm.
"""
from hyppo.independence import Hsic

from .regressors import fit_and_get_residuals


def get_edge_orientation(data, x, y, iters=20, method='gpr', verbose=False):
    """
    This is an ANM test of independence for the pairs between which
    a lot of correlation has been seen. If the test is repeated a sufficient
    number of times (100) the correct causal direction almost always comes
    out -- and in cases where it is not, it is enough to repeat the test
    an odd number of times (5) to see that the result Yeah that's right.

    Args:
        data (pandas.DataFrame): The dataset with the samples for the features
        x (str): the name of the source feature
        y (str): the name of the target feature
        iters (int, optional): Nr of repetitions of the test. Defaults to 100.
        method (str, optional): Can be 'gpr' or 'gam'. Defaults to 'gpr'.
        verbose (bool, optional): Verbosity. Defaults to False.

    Returns:
        int: Returns +1 if direction is x->y, or -1 if direction is x<-y
            Returns 0 if no direction can be set.
    """
    res_y = fit_and_get_residuals(
        data[x].values, data[y].values, method=method)
    res_x = fit_and_get_residuals(
        data[y].values, data[x].values, method=method)

    r1 = Hsic().test(res_y, data[x].values, reps=iters).pvalue
    r2 = Hsic().test(res_x, data[y].values, reps=iters).pvalue
    mark = "**" if r1 < 0.05 and r2 < 0.05 else ""
    if r1 > r2:
        if verbose:
            print(f" {x:>3s}-->{y:<3s}  "
                  f"[p({x:>3s}->{y:<3s}): {r1:8.6f};"
                  f" p({y:>3s}->{x:<3s}): {r2:8.6f}] {mark}")
        return +1
    elif r1 < r2:
        if verbose:
            print(f" {x:>3s}<--{y:<3s}  "
                  f"[p({x:>3s}->{y:<3s}): {r1:8.6f};"
                  f" p({y:>3s}->{x:<3s}): {r2:8.6f}] {mark}")
        return -1
    else:
        if verbose:
            print(f" {x:>3s}·-·{y:<3s}", end="")
            print(
                f"  [p({x:>3s}->{y:<3s}): {r1:8.6f}; p({y:>3s}->{x:<3s}): {r2:8.6f}]")
        return 0



import torch
import torch.nn as nn
import math


class LocallyConnected(nn.Module):
    """Local linear layer, i.e. Conv1dLocal() with filter size 1.

    Args:
        num_linear: num of local linear layers, i.e.
        in_features: m1
        out_features: m2
        bias: whether to include bias or not

    Shape:
        - Input: [n, d, m1]
        - Output: [n, d, m2]

    Attributes:
        weight: [d, m1, m2]
        bias: [d, m2]
    """

    def __init__(self, num_linear, input_features, output_features, bias=True):
        super(LocallyConnected, self).__init__()
        self.num_linear = num_linear
        self.input_features = input_features
        self.output_features = output_features

        self.weight = nn.Parameter(torch.Tensor(num_linear,
                                                input_features,
                                                output_features))
        if bias:
            self.bias = nn.Parameter(torch.Tensor(num_linear, output_features))
        else:
            # You should always register all possible parameters, but the
            # optional ones can be None if you want.
            self.register_parameter('bias', None)

        self.reset_parameters()

    @torch.no_grad()
    def reset_parameters(self):
        k = 1.0 / self.input_features
        bound = math.sqrt(k)
        nn.init.uniform_(self.weight, -bound, bound)
        if self.bias is not None:
            nn.init.uniform_(self.bias, -bound, bound)

    def forward(self, input: torch.Tensor):
        # [n, d, 1, m2] = [n, d, 1, m1] @ [1, d, m1, m2]
        out = torch.matmul(input.unsqueeze(dim=2), self.weight.unsqueeze(dim=0))
        out = out.squeeze(dim=2)
        if self.bias is not None:
            # [n, d, m2] += [d, m2]
            out += self.bias
        return out

    def extra_repr(self):
        # (Optional)Set the extra information about this module. You can test
        # it by printing an object of this class.
        return 'num_linear={}, in_features={}, out_features={}, bias={}'.format(
            self.num_linear, self.in_features, self.out_features,
            self.bias is not None
        )


def main():
    n, d, m1, m2 = 2, 3, 5, 7

    # numpy
    import numpy as np
    input_numpy = np.random.randn(n, d, m1)
    weight = np.random.randn(d, m1, m2)
    output_numpy = np.zeros([n, d, m2])
    for j in range(d):
        # [n, m2] = [n, m1] @ [m1, m2]
        output_numpy[:, j, :] = input_numpy[:, j, :] @ weight[j, :, :]

    # torch
    torch.set_default_dtype(torch.double)
    input_torch = torch.from_numpy(input_numpy)
    locally_connected = LocallyConnected(d, m1, m2, bias=False)
    locally_connected.weight.data[:] = torch.from_numpy(weight)
    output_torch = locally_connected(input_torch)

    # compare
    print(torch.allclose(output_torch, torch.from_numpy(output_numpy)))


if __name__ == '__main__':
    main()



"""
This Python version accomplishes the same task as the R function:

1. It uses `LassoCV` from scikit-learn to perform cross-validation and find the 
optimal regularization parameter (lambda in R, alpha in Python).
2. It then trains a final Lasso model using the optimal alpha.
3. The function returns a dictionary containing the fitted values, residuals, and 
the trained model.

Note that:
- The `cv.glmnet` in R is replaced by `LassoCV` in Python.
- The `glmnet` in R is replaced by `Lasso` in Python.
- In scikit-learn, the regularization parameter is called `alpha` instead of `lambda`.
- The `pars` parameter is kept for consistency, but it's not used in this 
implementation. You can extend the function to use additional parameters if needed.
- The cross-validation is set to 5-fold (you can adjust this if needed).
- A random state is set for reproducibility.

This Python version should provide equivalent functionality to the original R function.
"""
from typing import Dict, Any
from sklearn.linear_model import LassoCV, Lasso
import numpy as np


def train_lasso(X, y, pars=None) -> Dict[str, Any]:
    """_summary_

    Args:
        X (_type_): _description_
        y (_type_): _description_
        pars (_type_, optional): _description_. Defaults to None.

    Returns:
        _type_: _description_
    """
    # Perform cross-validation to find optimal lambda
    cv_model = LassoCV(cv=5, random_state=42)
    cv_model.fit(X, y)

    # Train the final model using the optimal lambda (alpha in sklearn)
    mod = Lasso(alpha=cv_model.alpha_)
    mod.fit(X, y)

    # Prepare results
    result = {
        'Yfit': mod.predict(X),
        'residuals': y - mod.predict(X),
        'model': mod
    }

    return result



import numpy as np
import pytest

from causalexplain.estimators.notears import utils


class FakeGraph:
    def __init__(self, adjacency):
        self.adjacency = np.array(adjacency)

    @classmethod
    def Weighted_Adjacency(cls, adjacency):
        return cls(adjacency)

    @classmethod
    def Adjacency(cls, adjacency):
        return cls(adjacency)

    def is_dag(self):
        # treat absence of cycles in upper triangle only for this stub
        return np.allclose(self.adjacency, np.tril(self.adjacency))

    def get_adjacency(self):
        class Data:
            def __init__(self, arr):
                self.data = arr.tolist()
        return Data(self.adjacency)

    def topological_sorting(self):
        return list(range(self.adjacency.shape[0]))

    def neighbors(self, node, mode=None):
        return list(np.where(self.adjacency[:, node] != 0)[0])


class FakeIG:
    IN = "IN"

    Graph = FakeGraph

    @staticmethod
    def Graph_Erdos_Renyi(*args, **kwargs):
        return FakeGraph(np.zeros((2, 2)))


def test_set_random_seed_reproducible():
    utils.set_random_seed(0)
    first = np.random.rand(3)
    utils.set_random_seed(0)
    second = np.random.rand(3)
    assert np.allclose(first, second)


def test_is_dag_with_fake_igraph(monkeypatch):
    monkeypatch.setattr(utils, "ig", FakeIG)
    dag_matrix = np.array([[0, 0], [1, 0]])
    assert utils.is_dag(dag_matrix)
    cyclic = np.array([[0, 1], [1, 0]])
    assert not utils.is_dag(cyclic)


def test_simulate_parameter_with_ranges():
    utils.set_random_seed(0)
    B = np.array([[0, 1], [0, 0]])
    W = utils.simulate_parameter(B, w_ranges=((-1.0, -0.5), (0.5, 1.0)))
    assert W.shape == B.shape
    assert np.all((W == 0) | (np.abs(W) >= 0.5))


def test_simulate_linear_sem_monkeypatched_ig(monkeypatch):
    monkeypatch.setattr(utils, "ig", FakeIG)
    monkeypatch.setattr(utils, "is_dag", lambda W: True)
    W = np.array([[0.0, 0.0], [1.0, 0.0]])
    X = utils.simulate_linear_sem(W, n=3, sem_type="gauss")
    assert X.shape == (3, 2)
    # second variable should depend on first through W[0,1]
    assert not np.allclose(X[:, 1], 0)


def test_simulate_nonlinear_sem_with_stubbed_graph(monkeypatch):
    monkeypatch.setattr(utils, "ig", FakeIG)
    B = np.array([[0, 0], [1, 0]])
    X = utils.simulate_nonlinear_sem(B, n=2, sem_type="mim", noise_scale=np.ones(2))
    assert X.shape == (2, 2)



# Generated by CodiumAI
import pytest

from causalexplain.explainability.regression_quality import RegQuality

# Tests that the method returns the correct indices of features that are both gamma
# and outliers for a list of scores with length > 1


def test_happy_path():
    # Arrange
    scores = [0.9186, 0.0916, 0.0594, 0.0119, 0.0206,
              0.0137, 0.0031, 0.002, 0.0016, 0.0009]
    expected_indices = {0}

    # Act
    result = RegQuality.predict(scores)

    # Assert
    assert result == expected_indices

# Tests that the method raises an exception when given an empty list of scores


def test_edge_case_empty_list():
    # Arrange
    scores = []
    # Assert
    result = RegQuality.predict(scores)
    assert result == set()

# Tests that the method returns an empty set when given a list of scores with length 1


def test_edge_case_single_score():
    # Arrange
    scores = [1.0]
    expected_indices = set()

    # Act
    result = RegQuality.predict(scores)

    # Assert
    assert result == expected_indices

# Tests that the method returns the correct indices of features that are both gamma
# and outliers when all scores are outliers


def test_other_case_all_outliers():
    # Arrange
    scores = [10.0, 20.0, 30.0, 40.0, 50.0]
    expected_indices = set()

    # Act
    result = RegQuality.predict(scores)

    # Assert
    assert result == expected_indices

# Tests that the method returns the correct indices of features that are both
# gamma and outliers when all scores are gamma


def test_other_case_all_gamma():
    # Arrange
    scores = [0.01, 0.002, 0.03, 0.4]
    expected_indices = {3}

    # Act
    result = RegQuality.predict(scores, verbose=True)

    # Assert
    assert result == expected_indices



import numpy as np
from sklearn.linear_model import LinearRegression


def train_linear(X, y, pars=None):
    """
    Train a linear regression model.

    Args:
        X (numpy.ndarray): Input features, shape (n_samples, n_features).
        y (numpy.ndarray): Target values, shape (n_samples, 1).
        pars (dict, optional): Additional parameters for the model. Defaults to None.

    Returns:
        dict: A dictionary containing:
            - 'Yfit' (numpy.ndarray): Predicted values, shape (n_samples, 1).
            - 'residuals' (numpy.ndarray): Residuals (y - y_pred), shape (n_samples, 1).
            - 'model' (LinearRegression): Fitted sklearn LinearRegression model.

    Note:
        The coefficients of the model can be accessed via the 'model' key in the 
        returned dictionary, specifically using `result['model'].coef_`.
    """
    if pars is None:
        pars = {}

    mod = LinearRegression().fit(X, y)

    result = {
        'Yfit': mod.predict(X).reshape(-1, 1),
        'residuals': (y - mod.predict(X)).reshape(-1, 1),
        'model': mod
    }

    # For coefficients, use mod.coef_
    return result



"""
Here are the main changes and explanations:

- We use `numpy` for array operations and `scipy.stats` for linear regression.
- The default parameter `pars` is set to `None` and then initialized if not provided.
- We use f-strings for string formatting in the print statement.
- The input `X` is converted to a numpy array.
- Instead of using `lm`, we use `scipy.stats.linregress` for linear regression.
- We manually add a constant term to `X` for the intercept in the regression.
- The p-values are extracted directly from the `linregress` result.
- The selection vector is updated using list slicing to exclude the k-th element.

Note that this Python version assumes that the input `X` is a 2D array-like object. 
The function will work similarly to the R version, but there might be slight 
differences in the exact numerical results due to different underlying implementations 
of the linear regression.
"""
import numpy as np
from scipy import stats


def selLm(X, pars=None, output=False, k: int = None):
    """_summary_

    Args:
        X (np.ndarray): a 2D numpy array with the variables
        pars (dict): Parameters
        output (bool, optional): _description_. Defaults to False.
        k (int, optional): The index of the variable

    Returns:
        _type_: _description_
    """
    if pars is None:
        pars = {'cut_off_p_val': 0.001}

    if output:
        print(f"Performing variable selection for variable {k}:")

    X = np.array(X)
    p = X.shape

    if p[1] > 1:
        sel_vec = [False] * p[1]
        x = np.delete(X, k, axis=1)
        y = X[:, k]

        # Perform linear regression
        X_with_const = np.column_stack((np.ones(len(x)), x))
        model = stats.linregress(X_with_const, y)

        # Calculate p-values for coefficients (excluding intercept)
        p_val_vec = model.pvalue[1:]

        # Update sel_vec based on p-values
        # We need to assign the boolean values to sel_vec, excluding the k-th element
        # First, we create a boolean array from p_val_vec
        significant = p_val_vec < pars['cut_off_p_val']

        # Then, we assign these values to sel_vec, skipping the k-th element
        sel_vec[:k] = significant[:k]
        sel_vec[k+1:] = significant[k:]
    else:
        sel_vec = []

    return sel_vec



from causalexplain.estimators.notears.locally_connected import LocallyConnected
from causalexplain.estimators.notears.lbfgsb_scipy import LBFGSBScipy
from causalexplain.estimators.notears.trace_expm import trace_expm
from causalexplain.estimators.notears import utils as ut
import torch
import torch.nn as nn
import numpy as np
import math


class NotearsMLP(nn.Module):
    def __init__(self, dims, bias=True):
        super(NotearsMLP, self).__init__()
        assert len(dims) >= 2
        assert dims[-1] == 1
        d = dims[0]
        self.dims = dims
        # fc1: variable splitting for l1
        self.fc1_pos = nn.Linear(d, d * dims[1], bias=bias)
        self.fc1_neg = nn.Linear(d, d * dims[1], bias=bias)
        self.fc1_pos.weight.bounds = self._bounds()
        self.fc1_neg.weight.bounds = self._bounds()
        # fc2: local linear layers
        layers = []
        for l in range(len(dims) - 2):
            layers.append(LocallyConnected(d, dims[l + 1], dims[l + 2], bias=bias))
        self.fc2 = nn.ModuleList(layers)

    def _bounds(self):
        d = self.dims[0]
        bounds = []
        for j in range(d):
            for m in range(self.dims[1]):
                for i in range(d):
                    if i == j:
                        bound = (0, 0)
                    else:
                        bound = (0, None)
                    bounds.append(bound)
        return bounds

    def forward(self, x):  # [n, d] -> [n, d]
        x = self.fc1_pos(x) - self.fc1_neg(x)  # [n, d * m1]
        x = x.view(-1, self.dims[0], self.dims[1])  # [n, d, m1]
        for fc in self.fc2:
            x = torch.sigmoid(x)  # [n, d, m1]
            x = fc(x)  # [n, d, m2]
        x = x.squeeze(dim=2)  # [n, d]
        return x

    def h_func(self):
        """Constrain 2-norm-squared of fc1 weights along m1 dim to be a DAG"""
        d = self.dims[0]
        fc1_weight = self.fc1_pos.weight - self.fc1_neg.weight  # [j * m1, i]
        fc1_weight = fc1_weight.view(d, -1, d)  # [j, m1, i]
        A = torch.sum(fc1_weight * fc1_weight, dim=1).t()  # [i, j]
        h = trace_expm(A) - d  # (Zheng et al. 2018)
        # A different formulation, slightly faster at the cost of numerical stability
        # M = torch.eye(d) + A / d  # (Yu et al. 2019)
        # E = torch.matrix_power(M, d - 1)
        # h = (E.t() * M).sum() - d
        return h

    def l2_reg(self):
        """Take 2-norm-squared of all parameters"""
        reg = 0.
        fc1_weight = self.fc1_pos.weight - self.fc1_neg.weight  # [j * m1, i]
        reg += torch.sum(fc1_weight ** 2)
        for fc in self.fc2:
            reg += torch.sum(fc.weight ** 2)
        return reg

    def fc1_l1_reg(self):
        """Take l1 norm of fc1 weight"""
        reg = torch.sum(self.fc1_pos.weight + self.fc1_neg.weight)
        return reg

    @torch.no_grad()
    def fc1_to_adj(self) -> np.ndarray:  # [j * m1, i] -> [i, j]
        """Get W from fc1 weights, take 2-norm over m1 dim"""
        d = self.dims[0]
        fc1_weight = self.fc1_pos.weight - self.fc1_neg.weight  # [j * m1, i]
        fc1_weight = fc1_weight.view(d, -1, d)  # [j, m1, i]
        A = torch.sum(fc1_weight * fc1_weight, dim=1).t()  # [i, j]
        W = torch.sqrt(A)  # [i, j]
        W = W.cpu().detach().numpy()  # [i, j]
        return W


class NotearsSobolev(nn.Module):
    def __init__(self, d, k):
        """d: num variables k: num expansion of each variable"""
        super(NotearsSobolev, self).__init__()
        self.d, self.k = d, k
        self.fc1_pos = nn.Linear(d * k, d, bias=False)  # ik -> j
        self.fc1_neg = nn.Linear(d * k, d, bias=False)
        self.fc1_pos.weight.bounds = self._bounds()
        self.fc1_neg.weight.bounds = self._bounds()
        nn.init.zeros_(self.fc1_pos.weight)
        nn.init.zeros_(self.fc1_neg.weight)
        self.l2_reg_store = None

    def _bounds(self):
        # weight shape [j, ik]
        bounds = []
        for j in range(self.d):
            for i in range(self.d):
                for _ in range(self.k):
                    if i == j:
                        bound = (0, 0)
                    else:
                        bound = (0, None)
                    bounds.append(bound)
        return bounds

    def sobolev_basis(self, x):  # [n, d] -> [n, dk]
        seq = []
        for kk in range(self.k):
            mu = 2.0 / (2 * kk + 1) / math.pi  # sobolev basis
            psi = mu * torch.sin(x / mu)
            seq.append(psi)  # [n, d] * k
        bases = torch.stack(seq, dim=2)  # [n, d, k]
        bases = bases.view(-1, self.d * self.k)  # [n, dk]
        return bases

    def forward(self, x):  # [n, d] -> [n, d]
        bases = self.sobolev_basis(x)  # [n, dk]
        x = self.fc1_pos(bases) - self.fc1_neg(bases)  # [n, d]
        self.l2_reg_store = torch.sum(x ** 2) / x.shape[0]
        return x

    def h_func(self):
        fc1_weight = self.fc1_pos.weight - self.fc1_neg.weight  # [j, ik]
        fc1_weight = fc1_weight.view(self.d, self.d, self.k)  # [j, i, k]
        A = torch.sum(fc1_weight * fc1_weight, dim=2).t()  # [i, j]
        h = trace_expm(A) - self.d  # (Zheng et al. 2018)
        # A different formulation, slightly faster at the cost of numerical stability
        # M = torch.eye(self.d) + A / self.d  # (Yu et al. 2019)
        # E = torch.matrix_power(M, self.d - 1)
        # h = (E.t() * M).sum() - self.d
        return h

    def l2_reg(self):
        reg = self.l2_reg_store
        return reg

    def fc1_l1_reg(self):
        reg = torch.sum(self.fc1_pos.weight + self.fc1_neg.weight)
        return reg

    @torch.no_grad()
    def fc1_to_adj(self) -> np.ndarray:
        fc1_weight = self.fc1_pos.weight - self.fc1_neg.weight  # [j, ik]
        fc1_weight = fc1_weight.view(self.d, self.d, self.k)  # [j, i, k]
        A = torch.sum(fc1_weight * fc1_weight, dim=2).t()  # [i, j]
        W = torch.sqrt(A)  # [i, j]
        W = W.cpu().detach().numpy()  # [i, j]
        return W


def squared_loss(output, target):
    n = target.shape[0]
    loss = 0.5 / n * torch.sum((output - target) ** 2)
    return loss


def dual_ascent_step(model, X, lambda1, lambda2, rho, alpha, h, rho_max):
    """Perform one step of dual ascent in augmented Lagrangian."""
    h_new = None
    optimizer = LBFGSBScipy(model.parameters())
    X_torch = torch.from_numpy(X)
    while rho < rho_max:
        def closure():
            optimizer.zero_grad()
            X_hat = model(X_torch)
            loss = squared_loss(X_hat, X_torch)
            h_val = model.h_func()
            penalty = 0.5 * rho * h_val * h_val + alpha * h_val
            l2_reg = 0.5 * lambda2 * model.l2_reg()
            l1_reg = lambda1 * model.fc1_l1_reg()
            primal_obj = loss + penalty + l2_reg + l1_reg
            primal_obj.backward()
            return primal_obj
        optimizer.step(closure)  # NOTE: updates model in-place
        with torch.no_grad():
            h_new = model.h_func().item()
        if h_new > 0.25 * h:
            rho *= 10
        else:
            break
    alpha += rho * h_new
    return rho, alpha, h_new


def notears_nonlinear(model: nn.Module,
                      X: np.ndarray,
                      lambda1: float = 0.,
                      lambda2: float = 0.,
                      max_iter: int = 100,
                      h_tol: float = 1e-8,
                      rho_max: float = 1e+16,
                      w_threshold: float = 0.3):
    rho, alpha, h = 1.0, 0.0, np.inf
    for _ in range(max_iter):
        rho, alpha, h = dual_ascent_step(model, X, lambda1, lambda2,
                                         rho, alpha, h, rho_max)
        if h <= h_tol or rho >= rho_max:
            break
    W_est = model.fc1_to_adj()
    W_est[np.abs(W_est) < w_threshold] = 0
    return W_est


def main():
    torch.set_default_dtype(torch.double)
    np.set_printoptions(precision=3)

    ut.set_random_seed(123)

    n, d, s0, graph_type, sem_type = 200, 5, 9, 'ER', 'mim'
    B_true = ut.simulate_dag(d, s0, graph_type)
    np.savetxt('W_true.csv', B_true, delimiter=',')

    X = ut.simulate_nonlinear_sem(B_true, n, sem_type)
    np.savetxt('X.csv', X, delimiter=',')

    model = NotearsMLP(dims=[d, 10, 1], bias=True)
    W_est = notears_nonlinear(model, X, lambda1=0.01, lambda2=0.01)
    assert ut.is_dag(W_est)
    np.savetxt('W_est.csv', W_est, delimiter=',')
    acc = ut.count_accuracy(B_true, W_est != 0)
    print(acc)


if __name__ == '__main__':
    main()



import networkx as nx
import pandas as pd
import pytest

from causalexplain.independence import effect


class FakeEstimate:
    def __init__(self, value):
        self.value = value


class FakeRefuter:
    def __init__(self, p_value):
        self.refutation_result = {"p_value": p_value}


class FakeCausalModel:
    def __init__(self, *, data, treatment, outcome, graph):
        self.data = data
        self.treatment = treatment
        self.outcome = outcome
        self.graph = graph
        self.calls = []

    def identify_effect(self, proceed_when_unidentifiable=True):
        self.calls.append(("identify", proceed_when_unidentifiable))
        return "identified"

    def estimate_effect(self, estimand, target_units="ate", method_name=None):
        self.calls.append(("estimate", estimand, target_units, method_name))
        return FakeEstimate(1.5)

    def refute_estimate(self, estimand, estimate, method_name=None):
        self.calls.append(("refute", estimand, estimate, method_name))
        return FakeRefuter(0.25)


def test_estimate_edge_uses_causal_model(monkeypatch):
    data = pd.DataFrame({"t": [0, 1, 0], "o": [1, 2, 3]})
    model_calls = {}

    def fake_model(**kwargs):
        cm = FakeCausalModel(**kwargs)
        model_calls["instance"] = cm
        return cm

    monkeypatch.setattr(effect, "CausalModel", fake_model)

    ate, pval = effect.estimate_edge(nx.DiGraph(), "t", "o", data, verbose=True)
    assert ate == 1.5 and pval == 0.25

    cm = model_calls["instance"]
    assert cm.calls[0][0] == "identify"
    assert cm.calls[1][0] == "estimate"
    assert cm.calls[2][0] == "refute"


def test_estimate_sets_attributes_and_respects_copy(monkeypatch):
    graph = nx.DiGraph()
    graph.add_edge("a", "b")
    data = pd.DataFrame({"a": [0, 1], "b": [1, 2]})

    monkeypatch.setattr(effect, "estimate_edge", lambda *args, **kwargs: ("ate", 0.1))

    modified = effect.estimate(graph, data, in_place=False)
    assert graph.get_edge_data("a", "b") == {}
    assert modified.get_edge_data("a", "b")["ate"] == "ate"
    assert modified.get_edge_data("a", "b")["refute_pval"] == 0.1



import numpy as np
import networkx as nx
from scipy.special import expit as sigmoid
import igraph as ig
import random


def set_random_seed(seed):
    random.seed(seed)
    np.random.seed(seed)


def is_dag(W):
    G = ig.Graph.Weighted_Adjacency(W.tolist())
    return G.is_dag()


def simulate_dag(d, s0, graph_type):
    """Simulate random DAG with some expected number of edges.

    Args:
        d (int): num of nodes
        s0 (int): expected num of edges
        graph_type (str): ER, SF, BP

    Returns:
        B (np.ndarray): [d, d] binary adj matrix of DAG
    """
    def _random_permutation(M):
        # np.random.permutation permutes first axis only
        P = np.random.permutation(np.eye(M.shape[0]))
        return P.T @ M @ P

    def _random_acyclic_orientation(B_und):
        return np.tril(_random_permutation(B_und), k=-1)

    def _graph_to_adjmat(G):
        return np.array(G.get_adjacency().data)

    if graph_type == 'ER':
        # Erdos-Renyi
        G_und = ig.Graph.Erdos_Renyi(n=d, m=s0)
        B_und = _graph_to_adjmat(G_und)
        B = _random_acyclic_orientation(B_und)
    elif graph_type == 'SF':
        # Scale-free, Barabasi-Albert
        G = ig.Graph.Barabasi(n=d, m=int(round(s0 / d)), directed=True)
        B = _graph_to_adjmat(G)
    elif graph_type == 'BP':
        # Bipartite, Sec 4.1 of (Gu, Fu, Zhou, 2018)
        top = int(0.2 * d)
        G = ig.Graph.Random_Bipartite(top, d - top, m=s0, directed=True, neimode=ig.OUT)
        B = _graph_to_adjmat(G)
    else:
        raise ValueError('unknown graph type')
    B_perm = _random_permutation(B)
    assert ig.Graph.Adjacency(B_perm.tolist()).is_dag()
    return B_perm


def simulate_parameter(B, w_ranges=((-2.0, -0.5), (0.5, 2.0))):
    """Simulate SEM parameters for a DAG.

    Args:
        B (np.ndarray): [d, d] binary adj matrix of DAG
        w_ranges (tuple): disjoint weight ranges

    Returns:
        W (np.ndarray): [d, d] weighted adj matrix of DAG
    """
    W = np.zeros(B.shape)
    S = np.random.randint(len(w_ranges), size=B.shape)  # which range
    for i, (low, high) in enumerate(w_ranges):
        U = np.random.uniform(low=low, high=high, size=B.shape)
        W += B * (S == i) * U
    return W


def simulate_linear_sem(W, n, sem_type, noise_scale=None):
    """Simulate samples from linear SEM with specified type of noise.

    For uniform, noise z ~ uniform(-a, a), where a = noise_scale.

    Args:
        W (np.ndarray): [d, d] weighted adj matrix of DAG
        n (int): num of samples, n=inf mimics population risk
        sem_type (str): gauss, exp, gumbel, uniform, logistic, poisson
        noise_scale (np.ndarray): scale parameter of additive noise, default all ones

    Returns:
        X (np.ndarray): [n, d] sample matrix, [d, d] if n=inf
    """
    def _simulate_single_equation(X, w, scale):
        """X: [n, num of parents], w: [num of parents], x: [n]"""
        if sem_type == 'gauss':
            z = np.random.normal(scale=scale, size=n)
            x = X @ w + z
        elif sem_type == 'exp':
            z = np.random.exponential(scale=scale, size=n)
            x = X @ w + z
        elif sem_type == 'gumbel':
            z = np.random.gumbel(scale=scale, size=n)
            x = X @ w + z
        elif sem_type == 'uniform':
            z = np.random.uniform(low=-scale, high=scale, size=n)
            x = X @ w + z
        elif sem_type == 'logistic':
            x = np.random.binomial(1, sigmoid(X @ w)) * 1.0
        elif sem_type == 'poisson':
            x = np.random.poisson(np.exp(X @ w)) * 1.0
        else:
            raise ValueError('unknown sem type')
        return x

    d = W.shape[0]
    if noise_scale is None:
        scale_vec = np.ones(d)
    elif np.isscalar(noise_scale):
        scale_vec = noise_scale * np.ones(d)
    else:
        if len(noise_scale) != d:
            raise ValueError('noise scale must be a scalar or has length d')
        scale_vec = noise_scale
    if not is_dag(W):
        raise ValueError('W must be a DAG')
    if np.isinf(n):  # population risk for linear gauss SEM
        if sem_type == 'gauss':
            # make 1/d X'X = true cov
            X = np.sqrt(d) * np.diag(scale_vec) @ np.linalg.inv(np.eye(d) - W)
            return X
        else:
            raise ValueError('population risk not available')
    # empirical risk
    G = ig.Graph.Weighted_Adjacency(W.tolist())
    ordered_vertices = G.topological_sorting()
    assert len(ordered_vertices) == d
    X = np.zeros([n, d])
    for j in ordered_vertices:
        parents = G.neighbors(j, mode=ig.IN)
        X[:, j] = _simulate_single_equation(X[:, parents], W[parents, j], scale_vec[j])
    return X


def simulate_nonlinear_sem(B, n, sem_type, noise_scale=None):
    """Simulate samples from nonlinear SEM.

    Args:
        B (np.ndarray): [d, d] binary adj matrix of DAG
        n (int): num of samples
        sem_type (str): mlp, mim, gp, gp-add
        noise_scale (np.ndarray): scale parameter of additive noise, default all ones

    Returns:
        X (np.ndarray): [n, d] sample matrix
    """
    def _simulate_single_equation(X, scale):
        """X: [n, num of parents], x: [n]"""
        z = np.random.normal(scale=scale, size=n)
        pa_size = X.shape[1]
        if pa_size == 0:
            return z
        if sem_type == 'mlp':
            hidden = 100
            W1 = np.random.uniform(low=0.5, high=2.0, size=[pa_size, hidden])
            W1[np.random.rand(*W1.shape) < 0.5] *= -1
            W2 = np.random.uniform(low=0.5, high=2.0, size=hidden)
            W2[np.random.rand(hidden) < 0.5] *= -1
            x = sigmoid(X @ W1) @ W2 + z
        elif sem_type == 'mim':
            w1 = np.random.uniform(low=0.5, high=2.0, size=pa_size)
            w1[np.random.rand(pa_size) < 0.5] *= -1
            w2 = np.random.uniform(low=0.5, high=2.0, size=pa_size)
            w2[np.random.rand(pa_size) < 0.5] *= -1
            w3 = np.random.uniform(low=0.5, high=2.0, size=pa_size)
            w3[np.random.rand(pa_size) < 0.5] *= -1
            x = np.tanh(X @ w1) + np.cos(X @ w2) + np.sin(X @ w3) + z
        elif sem_type == 'gp':
            from sklearn.gaussian_process import GaussianProcessRegressor
            gp = GaussianProcessRegressor()
            x = gp.sample_y(X, random_state=None).flatten() + z
        elif sem_type == 'gp-add':
            from sklearn.gaussian_process import GaussianProcessRegressor
            gp = GaussianProcessRegressor()
            x = sum([gp.sample_y(X[:, i, None], random_state=None).flatten()
                     for i in range(X.shape[1])]) + z
        else:
            raise ValueError('unknown sem type')
        return x

    d = B.shape[0]
    scale_vec = noise_scale if noise_scale is not None else np.ones(d)
    X = np.zeros([n, d])
    G = ig.Graph.Adjacency(B.tolist())
    ordered_vertices = G.topological_sorting()
    assert len(ordered_vertices) == d
    for j in ordered_vertices:
        parents = G.neighbors(j, mode=ig.IN)
        X[:, j] = _simulate_single_equation(X[:, parents], scale_vec[j])
    return X


def count_accuracy(B_true, B_est):
    """Compute various accuracy metrics for B_est.

    true positive = predicted association exists in condition in correct direction
    reverse = predicted association exists in condition in opposite direction
    false positive = predicted association does not exist in condition

    Args:
        B_true (np.ndarray): [d, d] ground truth graph, {0, 1}
        B_est (np.ndarray): [d, d] estimate, {0, 1, -1}, -1 is undirected edge in CPDAG

    Returns:
        fdr: (reverse + false positive) / prediction positive
        tpr: (true positive) / condition positive
        fpr: (reverse + false positive) / condition negative
        shd: undirected extra + undirected missing + reverse
        nnz: prediction positive
    """
    if (B_est == -1).any():  # cpdag
        if not ((B_est == 0) | (B_est == 1) | (B_est == -1)).all():
            raise ValueError('B_est should take value in {0,1,-1}')
        if ((B_est == -1) & (B_est.T == -1)).any():
            raise ValueError('undirected edge should only appear once')
    else:  # dag
        if not ((B_est == 0) | (B_est == 1)).all():
            raise ValueError('B_est should take value in {0,1}')
        if not is_dag(B_est):
            raise ValueError('B_est should be a DAG')
    d = B_true.shape[0]
    # linear index of nonzeros
    pred_und = np.flatnonzero(B_est == -1)
    pred = np.flatnonzero(B_est == 1)
    cond = np.flatnonzero(B_true)
    cond_reversed = np.flatnonzero(B_true.T)
    cond_skeleton = np.concatenate([cond, cond_reversed])
    # true pos
    true_pos = np.intersect1d(pred, cond, assume_unique=True)
    # treat undirected edge favorably
    true_pos_und = np.intersect1d(pred_und, cond_skeleton, assume_unique=True)
    true_pos = np.concatenate([true_pos, true_pos_und])
    # false pos
    false_pos = np.setdiff1d(pred, cond_skeleton, assume_unique=True)
    false_pos_und = np.setdiff1d(pred_und, cond_skeleton, assume_unique=True)
    false_pos = np.concatenate([false_pos, false_pos_und])
    # reverse
    extra = np.setdiff1d(pred, cond, assume_unique=True)
    reverse = np.intersect1d(extra, cond_reversed, assume_unique=True)
    # compute ratio
    pred_size = len(pred) + len(pred_und)
    cond_neg_size = 0.5 * d * (d - 1) - len(cond)
    fdr = float(len(reverse) + len(false_pos)) / max(pred_size, 1)
    tpr = float(len(true_pos)) / max(len(cond), 1)
    fpr = float(len(reverse) + len(false_pos)) / max(cond_neg_size, 1)
    # structural hamming distance
    pred_lower = np.flatnonzero(np.tril(B_est + B_est.T))
    cond_lower = np.flatnonzero(np.tril(B_true + B_true.T))
    extra_lower = np.setdiff1d(pred_lower, cond_lower, assume_unique=True)
    missing_lower = np.setdiff1d(cond_lower, pred_lower, assume_unique=True)
    shd = len(extra_lower) + len(missing_lower) + len(reverse)
    return {'fdr': fdr, 'tpr': tpr, 'fpr': fpr, 'shd': shd, 'nnz': pred_size}



def threshold_output(W, desired_edges=None, verbose=False):
    if desired_edges != None:
        # Implements binary search for acyclic graph with the desired number of edges
        ws = sorted([abs(i) for i in np.unique(W)])
        best = W
        done = False

        mid = int(len(ws)/2.0)
        floor = 0
        ceil = len(ws)-1

        while not done:
            cut = np.array([[0.0 if abs(i) <= ws[mid] else 1.0 for i in j] for j in W])
            g = nx.from_numpy_array(cut, create_using=nx.DiGraph())
            try:
                nx.find_cycle(g)
                floor = mid
                mid = int((ceil-mid)/2.0) + mid
                if mid == ceil or mid == floor:
                    done = True
            except:
                if nx.number_of_edges(g) == desired_edges:
                    best = cut
                    done = True
                elif nx.number_of_edges(g) >= desired_edges:
                    best = cut
                    floor = mid
                    mid = int((ceil-mid)/2.0) + mid
                    if mid == ceil or mid == floor:
                        done = True
                else:
                    best = cut
                    ceil = mid
                    mid = int((floor-mid)/2.0) + floor
                    if mid == ceil or mid == floor:
                        done = True
    else:
        ws = sorted([abs(i) for i in np.unique(W)])
        best = None

        for w in ws:
            cut = np.array([[0.0 if abs(i) <= w else 1.0 for i in j] for j in W])
            g = nx.from_numpy_array(cut, create_using=nx.DiGraph())
            try:
                nx.find_cycle(g)
            except:
                return cut
    return best


def generate_random_dag(num_nodes, num_edges, probabilistic=False, edge_coefficient_range=[0.5, 2.0]):
    adj_mat = np.zeros([num_nodes, num_nodes])

    ranking = np.array([i for i in range(num_nodes)])
    np.random.shuffle(ranking)

    if not probabilistic:
        for e_i in range(num_edges):
            open = np.argwhere(adj_mat == 0)
            open = list(filter(lambda x: ranking[x[0]] < ranking[x[1]], open))
            choice = np.random.choice([i for i in range(len(open))])
            adj_mat[open[choice][0]][open[choice][1]] = np.random.uniform(edge_coefficient_range[0], edge_coefficient_range[1]) * np.random.choice([-1.0, 1.0])


    else:
        p = num_edges/((num_nodes**2)/2.0)

        for i in range(num_nodes):
            for j in range(num_nodes):
                if ranking[i] < ranking[j]:
                    if np.random.rand() < p:
                           adj_mat[i][j] = np.random.uniform(edge_coefficient_range[0], edge_coefficient_range[1]) * np.random.choice([-1.0, 1.0])

    G_true = nx.from_numpy_array(adj_mat, create_using=nx.DiGraph())

    return adj_mat, G_true

def simulate_from_dag_lg(tam, n_sample, mean=0, variance=1):
    num_nodes = len(tam)

    def get_value(i, e):
        if values[i] == None:
            val = e[i]
            for j in range(num_nodes):
                if tam[j][i] != 0.0:
                    val += get_value(j, e) * tam[j][i]
            values[i] = val
            return val
        else:
            return values[i]

    simulation_data = []
    for i in range(n_sample):
        errors = np.random.normal(mean, variance, num_nodes)
        values = [None for _ in range(num_nodes)]
        for i in range(num_nodes):
            values[i] = get_value(i, errors)

        simulation_data.append(values)

    return simulation_data

def compare_graphs_undirected(true_graph, estimated_graph):
    num_edges = len(true_graph[np.where(true_graph != 0.0)])

    tam = np.array([[1 if x != 0.0 else 0.0 for x in y] for y in true_graph])
    tam_undir = tam + tam.T
    tam_undir = np.array([[1 if x != 0.0 else 0.0 for x in y] for y in tam_undir])
    tam_undir = np.triu(tam_undir)

    eam = np.array([[1 if x != 0.0 else 0.0 for x in y] for y in estimated_graph])
    eam_undir = eam + eam.T
    eam_undir = [[1 if x > 0 else 0 for x in y] for y in eam_undir]
    eam_undir = np.triu(eam_undir)

    tp = len(np.argwhere((tam_undir + eam_undir) == 2))
    fp = len(np.argwhere((tam_undir - eam_undir) < 0))
    tn = len(np.argwhere((tam_undir + eam_undir) == 0))
    fn = num_edges - tp

    return [tp, fp, tn, fn]

def compare_graphs_precision(x):
    if x[0] + x[1] == 0: return 0
    return float(x[0]) / float(x[0] + x[1])

def compare_graphs_recall(x):
    if x[0] + x[3] == 0: return 0
    return float(x[0]) / float(x[0] + x[3])

def compare_graphs_specificity(x):
    if x[2] + x[1] == 0: return 0
    return float(x[2]) / float(x[2] + x[1])



#
# Implementations of loss functions for DNNs.
# Reject using the one coming from CDT, because it is not working properly.
# Reject using the one at https://github.com/ZongxianLee/MMD_Loss.Pytorch
# Accepted using https://github.com/KevinMusgrave/pytorch-adapt
# Also valid: https://github.com/yiftachbeer/mmd_loss_pytorch/
#
# I keep the last one for simplicity. Results are the same.
#
import torch
import torch.nn as nn


class RBF(nn.Module):

    def __init__(self, n_kernels=5, mul_factor=2.0, bandwidth=None):
        super().__init__()
        self.bandwidth_multipliers = mul_factor ** (
            torch.arange(n_kernels) - n_kernels // 2)
        self.bandwidth = bandwidth

    def get_bandwidth(self, L2_distances):
        """
        Get the bandwidth of the RBF kernel.
        """
        if self.bandwidth is None:
            n_samples = L2_distances.shape[0]
            return L2_distances.data.sum() / (n_samples ** 2 - n_samples)

        return self.bandwidth

    def forward(self, X):
        L2_distances = torch.cdist(X, X) ** 2
        return torch.exp(-L2_distances[None, ...] / (self.get_bandwidth(L2_distances) * self.bandwidth_multipliers)[:, None, None]).sum(dim=0)


class MMDLoss(nn.Module):

    def __init__(self, kernel=RBF()):
        super().__init__()
        self.kernel = kernel

    def forward(self, X, Y):
        K = self.kernel(torch.vstack([X, Y]))

        X_size = X.shape[0]
        XX = K[:X_size, :X_size].mean()
        XY = K[:X_size, X_size:].mean()
        YY = K[X_size:, X_size:].mean()
        return XX - 2 * XY + YY



"""
This module contains the GBTRegressor class, which is a wrapper around the
GradientBoostingRegressor class from the scikit-learn library. The class
implements the fit, predict, and score methods to fit a separate model for
each feature in the dataframe, and predict and score the model for each feature
in the dataframe.

The class also implements a tune method to tune the hyperparameters of the model
using Optuna. The tune method uses the Objective class to define the objective
function for the hyperparameter optimization. The Objective class is a nested
class within the GBTRegressor class, and it defines the objective function for
the hyperparameter optimization. The class is designed to be used with the
Optuna library.

The module also contains a main function that can be used to run the GBTRegressor
class with the tune method. The main function takes the name of the experiment
as an argument, and loads the data and the reference graph for the experiment.
The main function then splits the data into train and test, and runs the tune
method to tune the hyperparameters of the model. The main function can be used
to run the GBTRegressor class with the tune method for any experiment.

The module can be run as a script to run the main function with the tune method
for a specific experiment. The experiment name is passed as an argument to the
script, and the main function is called with the experiment name as an argument.
The script can be used to run the GBTRegressor class with the tune method for
any experiment.

Example:

    $ python gbt.py rex_generated_linear_6

This will run the GBTRegressor class with the tune method for the experiment
'rex_generated_linear_6'.

The module can also be imported and used in other modules or scripts to run the
GBTRegressor class with the tune method for any experiment.

Example:

    from causalexplain.models.gbt import custom_main

    custom_main("rex_generated_linear_6")

"""

# pylint: disable=E1101:no-member, W0201:attribute-defined-outside-init, W0511:fixme
# pylint: disable=C0103:invalid-name
# pylint: disable=C0116:missing-function-docstring
# pylint: disable=R0913:too-many-arguments
# pylint: disable=R0914:too-many-locals, R0915:too-many-statements
# pylint: disable=W0106:expression-not-assigned, R1702:too-many-branches
# pylint: disable=W0102:dangerous-default-value


import inspect

import numpy as np
import optuna  # type: ignore
import pandas as pd
from mlforge.progbar import ProgBar  # type: ignore
from sklearn.ensemble import (GradientBoostingClassifier,
                              GradientBoostingRegressor)
from sklearn.metrics import f1_score
from sklearn.preprocessing import StandardScaler

from ..common import DEFAULT_HPO_TRIALS, utils
from ..explainability.hierarchies import Hierarchies


class GBTRegressor(GradientBoostingRegressor):

    random_state = 42

    def __init__(
            self,
            loss='squared_error',
            learning_rate=0.1,
            n_estimators=100,
            subsample=1.0,
            criterion='friedman_mse',
            min_samples_split=2,
            min_samples_leaf=1,
            min_weight_fraction_leaf=0.0,
            max_depth=3,
            min_impurity_decrease=0.0,
            init=None,
            random_state=42,
            max_features=None,
            # alpha=0.9,
            max_leaf_nodes=None,
            warm_start=False,
            validation_fraction=0.1,
            n_iter_no_change=None,
            tol=0.0001,
            ccp_alpha=0.0,
            correlation_th: float = None,
            verbose=False,
            silent=False,
            prog_bar=True,
            optuna_prog_bar=False):

        self.loss = loss
        self.learning_rate = learning_rate
        self.n_estimators = n_estimators
        self.subsample = subsample
        self.criterion = criterion
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_depth = max_depth
        self.min_impurity_decrease = min_impurity_decrease
        self.init = init
        self.random_state = random_state
        self.max_features = max_features
        # self.alpha = alpha
        self.max_leaf_nodes = max_leaf_nodes
        self.warm_start = warm_start
        self.validation_fraction = validation_fraction
        self.n_iter_no_change = n_iter_no_change
        self.tol = tol
        self.ccp_alpha = ccp_alpha
        self.correlation_th = correlation_th

        self.verbose = verbose
        self.silent = silent
        self.prog_bar = prog_bar
        self.optuna_prog_bar = optuna_prog_bar
        self.regressor = None
        self._estimator_name = 'gbt'
        self._estimator_class = GradientBoostingRegressor
        self._fit_desc = "Training GBTs"

    def fit(self, X):
        """
        Call the fit method of the parent class with every feature from the "X"
        dataframe as a target variable. This will fit a separate model for each
        feature in the dataframe.
        """
        self.n_features_in_ = X.shape[1]
        self.feature_names = utils.get_feature_names(X)
        self.feature_types = utils.get_feature_types(X)
        X = utils.cast_categoricals_to_int(X)
        self.regressor = dict()

        if self.correlation_th:
            self.corr_matrix = Hierarchies.compute_correlation_matrix(X)
            self.correlated_features = Hierarchies.compute_correlated_features(
                self.corr_matrix, self.correlation_th, self.feature_names,
                verbose=self.verbose)
            X_original = X.copy()

        # Who is calling me?
        try:
            curframe = inspect.currentframe()
            calframe = inspect.getouterframes(curframe, 2)
            caller_name = calframe[1][3]
            if caller_name == "__call__":
                caller_name = "HPO"
        except Exception:  # pylint: disable=broad-except
            caller_name = "unknown"

        if self.prog_bar and not self.verbose:
            pbar_name = f"({caller_name}) GBT_fit"
            pbar = ProgBar().start_subtask(pbar_name, len(self.feature_names))
        else:
            pbar = None

        for target_idx, target_name in enumerate(self.feature_names):
            # if correlation_th is not None then, remove features that are highly
            # correlated with the target, at each step of the loop
            if self.correlation_th is not None:
                X = X_original.copy()
                if len(self.correlated_features[target_name]) > 0:
                    X = X.drop(self.correlated_features[target_name], axis=1)
                    if self.verbose:
                        print("REMOVED CORRELATED FEATURES: ",
                              self.correlated_features[target_name])

            if self.feature_types[target_name] == 'categorical' or \
                    self.feature_types[target_name] == 'binary':
                self.loss = 'log_loss'
                gbt_model = GradientBoostingClassifier
            else:
                self.loss = 'squared_error'
                gbt_model = GradientBoostingRegressor

            self.regressor[target_name] = gbt_model(
                loss=self.loss,
                learning_rate=self.learning_rate,
                n_estimators=self.n_estimators,
                subsample=self.subsample,
                criterion=self.criterion,
                min_samples_split=self.min_samples_split,
                min_samples_leaf=self.min_samples_leaf,
                min_weight_fraction_leaf=self.min_weight_fraction_leaf,
                max_depth=self.max_depth,
                min_impurity_decrease=self.min_impurity_decrease,
                init=self.init,
                random_state=self.random_state,
                max_features=self.max_features,
                # alpha=self.alpha,
                verbose=False,
                max_leaf_nodes=self.max_leaf_nodes,
                warm_start=self.warm_start,
                validation_fraction=self.validation_fraction,
                n_iter_no_change=self.n_iter_no_change,
                tol=self.tol,
                ccp_alpha=self.ccp_alpha
            )

            self.regressor[target_name].fit(
                X.drop(target_name, axis=1), X[target_name])

            pbar.update_subtask(pbar_name, target_idx+1) if pbar else None

        pbar.remove(pbar_name) if pbar else None
        self.is_fitted_ = True
        return self

    def predict(self, X):
        """
        Call the predict method of the parent class with every feature from the "X"
        dataframe as a target variable. This will predict a separate value for each
        feature in the dataframe.
        """
        if not self.is_fitted_:
            raise ValueError(
                f"This {self.__class__.__name__} instance is not fitted yet."
                f"Call 'fit' with appropriate arguments before using this method.")
        y_pred = list()

        if self.correlation_th is not None:
            X_original = X.copy()

        for target_name in self.feature_names:
            if self.correlation_th is not None:
                X = X_original.drop(
                    self.correlated_features[target_name], axis=1)

            y_pred.append(
                self.regressor[target_name].predict(X.drop(target_name, axis=1)))

        return np.array(y_pred)

    def score(self, X):
        """
        Call the score method of the parent class with every feature from the "X"
        dataframe as a target variable. This will score a separate model for each
        feature in the dataframe.
        """
        if not self.is_fitted_:
            raise ValueError(
                f"This {self.__class__.__name__} instance is not fitted yet."
                f"Call 'fit' with appropriate arguments before using this method.")

        if self.correlation_th is not None:
            X_original = X.copy()

        scores = list()
        X_eval = utils.cast_categoricals_to_int(X)
        for target_name in self.feature_names:
            if self.correlation_th is not None:
                X_eval = X_original.drop(
                    self.correlated_features[target_name], axis=1)

            R2 = self.regressor[target_name].score(
                X_eval.drop(target_name, axis=1), X_eval[target_name])

            # Append 1.0 if R2 is negative, or 1.0 - R2 otherwise since we're
            # in the minimization mode of the error function.
            scores.append(1.0) if R2 < 0.0 else scores.append(1.0 - R2)

        self.scoring = np.array(scores)
        return self.scoring

    def tune(
        self,
        training_data: pd.DataFrame,
        test_data: pd.DataFrame,
        study_name: str = None,
        min_loss: float = 0.05,
        storage: str = "sqlite:///rex_tuning.db",
        load_if_exists: bool = True,
        n_trials: int = DEFAULT_HPO_TRIALS
    ):
        """
        Tune the hyperparameters of the model using Optuna.
        """
        class Objective:
            """
            A class to define the objective function for the hyperparameter optimization
            Some of the parameters for NNRegressor have been taken to default values to
            reduce the number of hyperparameters to optimize.

            Include this class in the hyperparameter optimization as follows:

            >>> study = optuna.create_study(direction='minimize',
            >>>                             study_name='study_name_here',
            >>>                             storage='sqlite:///db.sqlite3',
            >>>                             load_if_exists=True)
            >>> study.optimize(Objective(train_data, test_data), n_trials=100)

            The only dependency is you need to pass the train and test data to the class
            constructor. Tha class will build the data loaders for them from the
            dataframes.
            """

            def __init__(
                    self,
                    train_data,
                    test_data,
                    device='cpu',
                    prog_bar=True,
                    verbose=False):
                self.train_data = train_data
                self.test_data = test_data
                self.device = device
                self.random_state = GBTRegressor.random_state
                self.prog_bar = prog_bar
                self.verbose = verbose

            def __call__(self, trial):
                """
                This method is called by Optuna to evaluate the objective function.
                """
                # Define the model hyperparameters
                self.n_iter_no_change = 5
                self.tol = 0.0001

                # Define the hyperparameters to optimize
                self.learning_rate = trial.suggest_float(
                    "learning_rate", 0.001, 0.2)
                self.n_estimators = trial.suggest_int("n_estimators", 10, 1000)
                self.subsample = trial.suggest_float("subsample", 0.1, 1.0)
                self.min_samples_split = trial.suggest_int(
                    "min_samples_split", 2, 10)
                self.min_samples_leaf = trial.suggest_int(
                    "min_samples_leaf", 1, 10)
                self.min_weight_fraction_leaf = trial.suggest_float(
                    "min_weight_fraction_leaf", 0.0, 0.5)
                self.max_depth = trial.suggest_int("max_depth", 3, 20)
                self.max_leaf_nodes = trial.suggest_int(
                    "max_leaf_nodes", 10, 1000)
                self.min_impurity_decrease = trial.suggest_float(
                    "min_impurity_decrease", 0.0, 0.5)

                self.models = GBTRegressor(
                    learning_rate=self.learning_rate,
                    n_estimators=self.n_estimators,
                    subsample=self.subsample,
                    min_samples_split=self.min_samples_split,
                    min_samples_leaf=self.min_samples_leaf,
                    min_weight_fraction_leaf=self.min_weight_fraction_leaf,
                    max_depth=self.max_depth,
                    min_impurity_decrease=self.min_impurity_decrease,
                    random_state=self.random_state,
                    verbose=False,
                    max_leaf_nodes=self.max_leaf_nodes,
                    n_iter_no_change=self.n_iter_no_change,
                    tol=self.tol,
                    prog_bar=True & (not self.verbose) & (self.prog_bar),
                    silent=True)

                self.models.fit(self.train_data)

                # Now, measure the performance of the model with the test data.
                loss = []
                X_test = utils.cast_categoricals_to_int(self.test_data)
                for target_name in list(self.train_data.columns):
                    model = self.models.regressor[target_name]
                    # For regressors, this is R2, for classifiers this is accuracy
                    if model.__class__.__name__ == "GradientBoostingClassifier":
                        # Get the F1 score of the model
                        goodness_of_fit = f1_score(
                            X_test[target_name],
                            model.predict(X_test.drop(target_name, axis=1)))
                    elif model.__class__.__name__ == "GradientBoostingRegressor":
                        goodness_of_fit = model.score(
                            X_test.drop(target_name, axis=1), X_test[target_name])
                    else:
                        raise ValueError(
                            f"Model {model.__class__.__name__} is not supported."
                            f"Only GradientBoostingClassifier and"
                            f"GradientBoostingRegressor are supported.")

                    # Append 1.0 if R2 is negative, or 1.0 - R2 otherwise since we're
                    # in the minimization mode of the error function.
                    loss.append(1.0) if goodness_of_fit < 0.0 else loss.append(
                        1.0 - goodness_of_fit)

                return np.median(loss)

        # Callback function to stop the study if the loss is below a given threshold
        def callback(study: optuna.study.Study, trial: optuna.trial.FrozenTrial):
            if trial.value < min_loss or study.best_value < min_loss:
                study.stop()

        if self.verbose is False:
            optuna.logging.set_verbosity(optuna.logging.WARNING)

        # Create and run the HPO study
        study = optuna.create_study(
            direction='minimize', study_name=study_name, storage=storage,
            load_if_exists=load_if_exists)
        study.optimize(
            Objective(
                training_data, test_data, prog_bar=self.prog_bar, verbose=self.verbose),
            n_trials=n_trials,
            gc_after_trial=True,
            show_progress_bar=(self.optuna_prog_bar & (
                not self.silent) & (not self.verbose)),
            callbacks=[callback])

        # Capture the best hyperparameters and the minimum loss
        best_trials = sorted(study.best_trials, key=lambda x: x.values[0])
        self.best_params = best_trials[0].params
        self.min_tunned_loss = best_trials[0].values[0]

        if self.verbose and not self.silent:
            print(f"Best params (min loss:{self.min_tunned_loss:.6f}):")
            for k, v in self.best_params.items():
                print(f"\t{k:<15s}: {v}")

        regressor_args = {
            'learning_rate': self.best_params['learning_rate'],
            'n_estimators': self.best_params['n_estimators'],
            'subsample': self.best_params['subsample'],
            'min_samples_split': self.best_params['min_samples_split'],
            'min_samples_leaf': self.best_params['min_samples_leaf'],
            'min_weight_fraction_leaf': self.best_params['min_weight_fraction_leaf'],
            'max_depth': self.best_params['max_depth'],
            'max_leaf_nodes': self.best_params['max_leaf_nodes'],
            'min_impurity_decrease': self.best_params['min_impurity_decrease']
        }

        return regressor_args

    def tune_fit(
            self,
            X: pd.DataFrame,
            hpo_study_name: str = None,
            hpo_min_loss: float = 0.05,
            hpo_storage: str = 'sqlite:///rex_tuning.db',
            hpo_load_if_exists: bool = True,
            hpo_n_trials: int = DEFAULT_HPO_TRIALS):
        """
        Tune the hyperparameters of the model using Optuna, and the fit the model
        with the best parameters.
        """
        # split X into train and test
        train_data = X.sample(frac=0.9, random_state=self.random_state)
        test_data = X.drop(train_data.index)

        # tune the model
        regressor_args = self.tune(
            train_data, test_data, n_trials=hpo_n_trials, study_name=hpo_study_name,
            min_loss=hpo_min_loss, storage=hpo_storage,
            load_if_exists=hpo_load_if_exists)

        if self.verbose and not self.silent:
            print(f"Best params (min loss:{self.min_tunned_loss:.6f}):")
            for k, v in regressor_args.items():
                print(f"\t{k:<15s}: {v}")

        # Set the object parameters to the best parameters found.
        for k, v in regressor_args.items():
            setattr(self, k, v)

        # Fit the model with the best parameters.
        self.fit(train_data)


#
# Main function
#

def custom_main(experiment_name='custom_rex', score: bool = False, tune: bool = False):
    from causalexplain.common import utils
    path = "/Users/renero/phd/data/RC3/"
    output_path = "/Users/renero/phd/output/RC3/"

    ref_graph = utils.graph_from_dot_file(f"{path}{experiment_name}.dot")

    data = pd.read_csv(f"{path}{experiment_name}.csv")
    scaler = StandardScaler()
    data = pd.DataFrame(scaler.fit_transform(data), columns=data.columns)

    # Split the dataframe into train and test
    train = data.sample(frac=0.9, random_state=42)
    test = data.drop(train.index)

    if score:
        rex = utils.load_experiment(f"{experiment_name}", output_path)
        rex.is_fitted_ = True
        print(f"Loaded experiment {experiment_name}")
        rex.models.score(test)
    elif tune:
        gbt = GBTRegressor(silent=True, prog_bar=False)  # verbose=True)
        gbt.tune_fit(train, hpo_study_name=experiment_name, hpo_n_trials=100)
        print(gbt.score(test))


if __name__ == "__main__":
    custom_main("rex_generated_linear_6", tune=True)



"""
This Python version of `selGamBoost` follows the structure and logic of the original
R function. Here are some key points about the translation:

1. We import `numpy` for array operations and assume that `train_GAMboost` is
imported from a separate file.
2. The function signature remains similar, with default values for `pars` and `output`.
3. R's matrix indexing is replaced with NumPy array indexing.
4. The `cat` function for output is replaced with Python's `print` function.
5. List comprehensions and NumPy functions are used to replace some R-specific
    operations.
6. The `xselect()` method is assumed to exist in the model returned by `train_GAMboost`.
You may need to adjust this based on the actual implementation.
7. The boolean indexing and selection logic is adapted to work with NumPy arrays.

Note that this translation assumes that the `train_GAMboost` function in Python returns
an object with similar properties to its R counterpart. You may need to adjust the
code further based on the exact implementation of `train_GAMboost` in Python.
"""
import numpy as np
from .train_GAMboost import train_GAMboost


def selGamBoost(X, pars=None, output=False, k=None):
    """_summary_

    Args:
        X (_type_): _description_
        pars (_type_, optional): _description_. Defaults to None.
        output (bool, optional): _description_. Defaults to False.
        k (_type_, optional): _description_. Defaults to None.

    Returns:
        _type_: _description_
    """
    if pars is None:
        pars = {'atLeastThatMuchSelected': 0.02, 'atMostThatManyNeighbors': 10}

    if output:
        print(f"Performing variable selection for variable {k}:")

    result = {}
    X = np.array(X)
    p = X.shape

    if p[1] > 1:
        selVec = [False] * p[1]
        X_without_k = np.delete(X, k, axis=1)
        X_k = X[:, k]
        modfitGam = train_GAMboost(X_without_k, X_k, pars)

        # Replace the xselect() call with feature_importances_
        feature_importances = modfitGam['model'].feature_importances_
        # Indices of features sorted by importance
        cc = np.argsort(feature_importances)[::-1]
        if output:
            print("The following variables")
            print(cc)

        nstep = len(feature_importances)
        howOftenSelected = feature_importances[cc]

        if output:
            print("... have been selected that many times:")
            print(howOftenSelected)

        howOftenSelectedSorted = sorted(howOftenSelected, reverse=True)

        if sum(np.array(howOftenSelected) > pars['atLeastThatMuchSelected']) \
                > pars['atMostThatManyNeighbors']:
            cc = cc[np.array(howOftenSelected) >
                    howOftenSelectedSorted[pars['atMostThatManyNeighbors']]]
        else:
            cc = cc[np.array(howOftenSelected) >
                    pars['atLeastThatMuchSelected']]

        if output:
            print("We finally choose as possible parents:")
            print(cc)
            print()

        tmp = [False] * (p[1] - 1)
        for i in cc:
            tmp[i] = True
        selVec[:k] + tmp + selVec[k+1:]
    else:
        selVec = []

    return selVec



import numpy as np
import pytest

import causalexplain.independence.hsic as hsic_mod
from causalexplain.independence.hsic import HSIC


def test_hsic_fit_wraps_independence_test(monkeypatch):
    called = {}

    def fake_independence_test(X, Y, conditioned_on=None, method=None):
        called["args"] = (X.copy(), Y.copy(), conditioned_on, method)
        return 0.2

    monkeypatch.setattr(hsic_mod, "independence_test", fake_independence_test)
    h = HSIC()
    result = h.fit(np.array([1, 2]), np.array([2, 3]))

    assert called["args"][2] is None and called["args"][3] == "kernel"
    assert result.independence
    assert pytest.approx(result.p_value) == 0.2



import numpy as np
import networkx as nx
import pytest
from causalexplain.metrics.compare_graphs import (
    Metrics, _adjacency, _aupr, _binary_adj_matrix, _conf_mat,
    _conf_mat_directed, _conf_mat_undirected, _confusion_matrix, _f1,
    _is_weighted, _intersect_matrices, _negative, _positive, _precision,
    _recall, _SHD, _shallow_copy, _weighted_adjacency, evaluate_graph
)


@pytest.fixture(autouse=True)
def _isolated_matplotlib_cache(tmp_path, monkeypatch):
    monkeypatch.setenv("MPLCONFIGDIR", str(tmp_path))


@pytest.fixture
def sid_statistics():
    return {"sid": 2.0, "sidLowerBound": 1.0, "sidUpperBound": 3.0}


@pytest.fixture
def metrics_instance(sid_statistics):
    return Metrics(
        Tp=3,
        Tn=5,
        Fn=2,
        Fp=1,
        precision=0.75,
        recall=0.6,
        aupr=0.65,
        f1=0.6666666667,
        shd=2,
        sid=sid_statistics,
    )


@pytest.fixture
def directed_graph_pair():
    truth = nx.DiGraph()
    truth.add_edges_from([("a", "b"), ("b", "c")])
    est = nx.DiGraph()
    est.add_edges_from([("a", "b"), ("c", "b")])
    feature_names = ["a", "b", "c"]
    return truth, est, feature_names


@pytest.fixture
def undirected_graph_pair():
    truth = nx.Graph()
    truth.add_edges_from([("a", "b"), ("b", "c")])
    est = nx.Graph()
    est.add_edges_from([("a", "b")])
    feature_names = ["a", "b", "c"]
    return truth, est, feature_names


def test_metrics_initialization_and_to_dict(metrics_instance, sid_statistics):
    result = metrics_instance.to_dict()
    assert result["Tp"] == 3
    assert result["Tn"] == 5
    assert result["Fn"] == 2
    assert result["Fp"] == 1
    assert metrics_instance.ishd == pytest.approx(1 - (2 / 3))
    assert metrics_instance.sid_lower == sid_statistics["sidLowerBound"]
    assert metrics_instance.sid_upper == sid_statistics["sidUpperBound"]


def test_metrics_str_and_matplotlib_repr(metrics_instance):
    text = str(metrics_instance)
    assert "Precision: 0.75" in text
    assert "Recall...: 0.6" in text
    assert "[1.0..3.0]" in text
    plt_text = metrics_instance.matplotlib_repr()
    assert "> Precision: 0.75" in plt_text
    assert "> SID......: 2.0" in plt_text


def test_metrics_handles_zero_errors_in_str():
    metrics = Metrics(
        Tp=1,
        Tn=0,
        Fn=0,
        Fp=0,
        precision=1.0,
        recall=1.0,
        aupr=1.0,
        f1=1.0,
        shd=0,
        sid={"sid": 0.0, "sidLowerBound": 0.0, "sidUpperBound": 0.0},
    )
    sid_line = str(metrics).splitlines()[-1]
    assert "[" not in sid_line
    assert metrics.ishd == 0.0
    assert "> SID......: 0.0" in metrics.matplotlib_repr()


def test_shallow_copy_isolated_from_original():
    graph = nx.DiGraph()
    graph.add_edges_from([("a", "b"), ("b", "c")])
    copy = _shallow_copy(graph)
    assert isinstance(copy, nx.DiGraph)
    assert set(copy.edges()) == {("a", "b"), ("b", "c")}
    copy.add_edge("c", "a")
    assert ("c", "a") not in graph.edges()


def test_evaluate_graph_returns_none_when_truth_missing():
    predicted = nx.DiGraph()
    predicted.add_edge("a", "b")
    assert evaluate_graph(None, predicted) is None


def test_evaluate_graph_computes_metrics_for_directed_graphs(directed_graph_pair):
    truth, est, _ = directed_graph_pair
    metrics = evaluate_graph(truth, est)
    assert metrics.Tp == 1
    assert metrics.Tn == 3
    assert metrics.Fn == 1
    assert metrics.Fp == 1
    assert metrics.shd == 2
    assert metrics.sid == 3
    assert metrics.aupr == pytest.approx(5 / 9)


def test_evaluate_graph_handles_missing_nodes_and_feature_names():
    """Prediction graphs missing nodes must still produce valid metrics."""
    truth = nx.DiGraph()
    truth.add_edges_from([("a", "b"), ("b", "c")])
    truth.add_node("d")
    predicted = nx.DiGraph()
    predicted.add_edge("a", "b")
    metrics = evaluate_graph(truth, predicted,
                             feature_names=["a", "b", "c", "d"])
    assert metrics.Tp == 1
    assert metrics.Tn == 10
    assert metrics.Fn == 1
    assert metrics.Fp == 0
    assert metrics.sid == 2
    assert metrics.aupr == pytest.approx(0.78125)


def test_evaluate_graph_double_for_anticausal_toggle():
    """SHD drops when reversed edges are not double-counted."""
    truth = nx.DiGraph()
    truth.add_edge("a", "b")
    predicted = nx.DiGraph()
    predicted.add_edge("b", "a")
    default_metrics = evaluate_graph(truth, predicted)
    relaxed_metrics = evaluate_graph(
        truth, predicted, double_for_anticausal=False)
    assert default_metrics.shd == 2
    assert relaxed_metrics.shd == 1


def test_evaluate_graph_mismatched_directedness_raises():
    """Different graph types currently trigger an assertion."""
    truth = nx.Graph()
    truth.add_edge("a", "b")
    predicted = nx.DiGraph()
    predicted.add_edge("a", "b")
    with pytest.raises(AssertionError):
        evaluate_graph(truth, predicted)


def test_conf_mat_directed_wrapper(directed_graph_pair):
    truth, est, feature_names = directed_graph_pair
    assert _conf_mat(truth, est, feature_names) == (1, 3, 1, 1)


def test_conf_mat_undirected_wrapper(undirected_graph_pair):
    truth, est, feature_names = undirected_graph_pair
    assert _conf_mat(truth, est, feature_names) == (1, 1, 0, 1)


def test_confusion_matrix_directed_branch():
    graph = nx.DiGraph()
    graph.add_nodes_from(["a", "b"])
    metrics = {
        "_Gm": np.array([[0, 1], [0, 0]], dtype=int),
        "_gm": np.array([[0, 1], [0, 0]], dtype=int),
    }
    assert _confusion_matrix(graph, metrics) == (1, 1, 0, 0)


def test_confusion_matrix_undirected_branch():
    graph = nx.Graph()
    graph.add_nodes_from(["a", "b"])
    metrics = {
        "_Gm": np.array([[0, 1], [1, 0]], dtype=int),
        "_gm": np.array([[0, 0], [0, 0]], dtype=int),
    }
    assert _confusion_matrix(graph, metrics) == (0, 0, -1, 0)


def test_confusion_matrix_rejects_unknown_graph_type():
    """The helper validates that the graph is a supported NetworkX type."""
    metrics = {"_Gm": np.zeros((1, 1)), "_gm": np.zeros((1, 1))}
    with pytest.raises(TypeError):
        _confusion_matrix(object(), metrics)


def test_shd_counts_directional_disagreements_twice():
    metrics = {
        "_Gm": np.array([[0, 1], [0, 0]]),
        "_gm": np.array([[0, 0], [1, 0]]),
        "_double_for_anticausal": True,
    }
    assert _SHD(metrics) == 2
    metrics["_double_for_anticausal"] = False
    assert _SHD(metrics) == 1


def test_shd_zero_for_identical_matrices():
    metrics = {
        "_Gm": np.zeros((2, 2), dtype=int),
        "_gm": np.zeros((2, 2), dtype=int),
        "_double_for_anticausal": True,
    }
    assert _SHD(metrics) == 0


# Two identical graphs are compared.
def test_identical_graphs():
    truth = nx.DiGraph()
    truth.add_edges_from([("a", "b"), ("b", "c")])
    est = nx.DiGraph()
    est.add_edges_from([("a", "b"), ("b", "c")])
    feature_names = ["a", "b", "c"]

    tp, tn, fn, fp = _conf_mat_directed(truth, est, feature_names)
    assert tp == 2
    assert tn == 4
    assert fp == 0
    assert fn == 0

# Two graphs, same nodes, with one edge different are compared.


def test_one_edge_different():
    truth = nx.DiGraph()
    truth.add_edges_from([("a", "b"), ("b", "c")])
    est = nx.DiGraph()
    est.add_edges_from([("a", "b"), ("c", "a")])
    feature_names = ["a", "b", "c"]

    tp, tn, fn, fp = _conf_mat_directed(truth, est, feature_names)
    assert tp == 1
    assert tn == 3
    assert fp == 1
    assert fn == 1

# Two graphs with different edges are compared.


def test_different_edges():
    truth = nx.DiGraph()
    truth.add_edge("a", "b")
    truth.add_edge("b", "c")
    est = nx.DiGraph()
    est.add_edge("a", "b")
    est.add_edge("a", "c")
    feature_names = ["a", "b", "c"]

    tp, tn, fn, fp = _conf_mat_directed(truth, est, feature_names)

    assert tp == 1
    assert fp == 1
    assert tn == 3
    assert fn == 1

# Two graphs with different nodes are compared.


def test_different_nodes_est_more_connections():
    truth = nx.DiGraph()
    truth.add_edge("a", "b")
    truth.add_edge("b", "c")
    truth.add_node("d")
    est = nx.DiGraph()
    est.add_edge("a", "b")
    est.add_edge("b", "c")
    est.add_edge("c", "d")
    feature_names = ["a", "b", "c", "d"]

    tp, tn, fn, fp = _conf_mat_directed(truth, est, feature_names)

    assert tp == 2
    assert tn == 9
    assert fp == 0
    assert fn == 1

# Two graphs with different nodes (truth has more connection) are compared.


def test_different_nodes_truth_more_connections():
    truth = nx.DiGraph()
    truth.add_edge("a", "b")
    truth.add_edge("b", "c")
    truth.add_edge("c", "d")
    est = nx.DiGraph()
    est.add_edge("a", "b")
    est.add_edge("b", "c")
    est.add_node("d")
    feature_names = ["a", "b", "c", "d"]

    tp, tn, fn, fp = _conf_mat_directed(truth, est, feature_names)

    assert tp == 2
    assert fp == 1
    assert tn == 9
    assert fn == 0


# Two graphs with one node and no edges are compared.
def test_one_node_no_edges():
    truth = nx.DiGraph()
    truth.add_node("a")
    est = nx.DiGraph()
    est.add_node("a")
    feature_names = ["a"]

    tp, tn, fn, fp = _conf_mat_directed(truth, est, feature_names)

    assert tp == 0
    assert fp == 0
    assert tn == 0
    assert fn == 0

# Two graphs with one node and a self-loop are compared.


def test_one_node_self_loop():
    truth = nx.DiGraph()
    truth.add_edge("a", "a")
    est = nx.DiGraph()
    est.add_edge("a", "a")
    feature_names = ["a"]

    tp, tn, fn, fp = _conf_mat_directed(truth, est, feature_names)

    assert tp == 1
    assert fp == 0
    assert tn == 0
    assert fn == 0

# Two graphs with multiple nodes and no edges are compared.


def test_multiple_nodes_no_edges():
    truth = nx.DiGraph()
    truth.add_nodes_from(["a", "b", "c"])
    est = nx.DiGraph()
    est.add_nodes_from(["a", "b", "c"])
    feature_names = ["a", "b", "c"]

    tp, tn, fn, fp = _conf_mat_directed(truth, est, feature_names)

    assert tp == 0
    assert fp == 0
    assert tn == 6
    assert fn == 0


def test_unweighted_graph():
    G = nx.Graph()
    G.add_edge("a", "b")
    G.add_edge("b", "c")
    assert not _is_weighted(G)


def test_weighted_graph():
    G = nx.Graph()
    G.add_edge("a", "b", weight=1.0)
    G.add_edge("b", "c", weight=2.0)
    assert _is_weighted(G)


def test_mixed_graph():
    G = nx.Graph()
    G.add_edge("a", "b")
    G.add_edge("b", "c", weight=2.0)
    assert _is_weighted(G)


# Tests for the _binary_adj_matrix

def test_binary_adj_matrix_default_threshold():
    G = nx.DiGraph()
    G.add_edge("a", "b", weight=0.5)
    G.add_edge("b", "c", weight=0.2)
    G.add_edge("c", "a", weight=0.0)
    expected_matrix = np.array([[0, 1, 0], [0, 0, 1], [1, 0, 0]])
    result_matrix = _binary_adj_matrix(G)
    np.testing.assert_array_equal(result_matrix, expected_matrix)


def test_binary_adj_matrix_with_threshold():
    G = nx.DiGraph()
    G.add_edge("a", "b", weight=0.5)
    G.add_edge("b", "c", weight=0.2)
    G.add_edge("c", "a", weight=0.0)
    expected_matrix = np.array([[0, 1, 0], [0, 0, 0], [0, 0, 0]])
    result_matrix = _binary_adj_matrix(G, threshold=0.3)
    np.testing.assert_array_equal(result_matrix, expected_matrix)


def test_binary_adj_matrix_with_absolute():
    G = nx.DiGraph()
    G.add_edge("a", "b", weight=-0.5)
    G.add_edge("b", "c", weight=0.2)
    G.add_edge("c", "a", weight=-0.4)
    expected_matrix = np.array([[0, 1, 0], [0, 0, 1], [1, 0, 0]])
    result_matrix = _binary_adj_matrix(G, absolute=True)
    np.testing.assert_array_equal(result_matrix, expected_matrix)


def test_binary_adj_matrix_with_order():
    G = nx.DiGraph()
    G.add_edge("a", "b", weight=0.5)
    G.add_edge("b", "c", weight=0.2)
    G.add_edge("c", "a", weight=0.0)
    expected_matrix = np.array([[0, 0, 1], [1, 0, 0], [0, 1, 0]])
    result_matrix = _binary_adj_matrix(G, order=["c", "b", "a"])
    np.testing.assert_array_equal(result_matrix, expected_matrix)


# Test for the _adjacency function

def test_adjacency_unweighted_graph():
    G = nx.Graph()
    G.add_edge("a", "b")
    G.add_edge("b", "c")
    expected_matrix = np.array([[0, 1, 0], [1, 0, 1], [0, 1, 0]])
    result = _adjacency(G)
    np.testing.assert_array_equal(result, expected_matrix)


def test_adjacency_weighted_graph_with_threshold():
    G = nx.Graph()
    G.add_edge("a", "b", weight=0.5)
    G.add_edge("b", "c", weight=0.8)
    G.add_edge("a", "c", weight=0.3)
    expected_matrix = np.array([[0, 1, 0], [0, 0, 1], [0, 0, 0]])
    result = _adjacency(G, threshold=0.4)
    np.testing.assert_array_equal(result, expected_matrix)


def test_adjacency_with_unoriented_order():
    G = nx.Graph()
    G.add_edge("a", "b")
    G.add_edge("b", "c")
    order = ["b", "a", "c"]
    expected_matrix = np.array([[0, 1, 1], [1, 0, 0], [1, 0, 0]])
    result = _adjacency(G, order=order)
    np.testing.assert_array_equal(result, expected_matrix)


def test_adjacency_with_oriented_order():
    G = nx.DiGraph()
    G.add_edge("a", "b")
    G.add_edge("b", "c")
    order = ["b", "a", "c"]
    expected_matrix = np.array([[0, 0, 1], [1, 0, 0], [0, 0, 0]])
    result = _adjacency(G, order=order)
    np.testing.assert_array_equal(result, expected_matrix)

# Tests for the _weighted_adjacency function


def test_weighted_adjacency_unweighted_graph():
    G = nx.Graph()
    G.add_edge("a", "b")
    G.add_edge("b", "c")
    expected_matrix = np.array([[0, 1, 0], [1, 0, 1], [0, 1, 0]])
    result = _weighted_adjacency(G)
    np.testing.assert_array_equal(result, expected_matrix)


def test_weighted_adjacency_with_threshold():
    G = nx.Graph()
    G.add_edge("a", "b", weight=0.5)
    G.add_edge("b", "c", weight=0.8)
    G.add_edge("a", "c", weight=0.3)
    expected_matrix = np.array([[0, 0.5, 0], [0.5, 0, 0.8], [0, 0.8, 0]])
    result = _weighted_adjacency(G, threshold=0.4)
    np.testing.assert_array_equal(result, expected_matrix)


def test_weighted_adjacency_with_absolute():
    G = nx.Graph()
    G.add_edge("a", "b", weight=-0.5)
    G.add_edge("b", "c", weight=0.8)
    G.add_edge("a", "c", weight=-0.3)
    expected_matrix = np.array([[0, 0.5, 0], [0.5, 0, 0.8], [0, 0.8, 0]])
    result = _weighted_adjacency(G, threshold=0.4, absolute=True)
    np.testing.assert_array_equal(result, expected_matrix)


def test_weighted_adjacency_with_order():
    G = nx.Graph()
    G.add_edge("a", "b", weight=0.5)
    G.add_edge("b", "c", weight=0.8)
    order = ["b", "a", "c"]
    expected_matrix = np.array([[0, 0.5, 0.8], [0.5, 0, 0], [0.8, 0, 0]])
    result = _weighted_adjacency(G, order=order)
    np.testing.assert_array_equal(result, expected_matrix)


def test_weighted_adjacency_oriented_with_order():
    G = nx.DiGraph()
    G.add_edge("a", "b", weight=0.5)
    G.add_edge("b", "c", weight=0.8)
    order = ["b", "a", "c"]
    expected_matrix = np.array([[0, 0, 0.8], [0.5, 0, 0], [0, 0, 0]])
    result = _weighted_adjacency(G, order=order)
    np.testing.assert_array_equal(result, expected_matrix)

# Tests for the _intersect_matrices function


def test_intersect_matrices_identical():
    mat1 = np.array([[1, 0], [0, 1]])
    mat2 = np.array([[1, 0], [0, 1]])
    expected_result = np.array([[1, 0], [0, 1]])
    result = _intersect_matrices(mat1, mat2)
    np.testing.assert_array_equal(result, expected_result)


def test_intersect_matrices_different():
    mat1 = np.array([[1, 0], [0, 1]])
    mat2 = np.array([[0, 1], [1, 0]])
    expected_result = np.array([[0, 0], [0, 0]])
    result = _intersect_matrices(mat1, mat2)
    np.testing.assert_array_equal(result, expected_result)


def test_intersect_matrices_different_shapes():
    mat1 = np.array([[1, 0, 1], [0, 1, 0]])
    mat2 = np.array([[1, 0], [0, 1]])
    try:
        _intersect_matrices(mat1, mat2)
    except ValueError as e:
        assert str(e) == "Both graphs must have the same number of nodes"

# Tests for the _positive function


def test_positive_all_positive_values():
    matrix = np.array([[1, 2], [3, 4]])
    expected_result = np.array([[1, 2], [3, 4]])
    result = _positive(matrix)
    np.testing.assert_array_equal(result, expected_result)


def test_positive_all_negative_values():
    matrix = np.array([[-1, -2], [-3, -4]])
    expected_result = np.array([[0, 0], [0, 0]])
    result = _positive(matrix)
    np.testing.assert_array_equal(result, expected_result)


def test_positive_mixed_values():
    matrix = np.array([[1, -2], [-3, 4]])
    expected_result = np.array([[1, 0], [0, 4]])
    result = _positive(matrix)
    np.testing.assert_array_equal(result, expected_result)


def test_positive_zero_values():
    matrix = np.array([[0, 0], [0, 0]])
    expected_result = np.array([[0, 0], [0, 0]])
    result = _positive(matrix)
    np.testing.assert_array_equal(result, expected_result)

# Tests for the _negative function


def test_negative_all_positive_values():
    matrix = np.array([[1, 2], [3, 4]])
    expected_result = np.array([[0, 0], [0, 0]])
    result = _negative(matrix)
    np.testing.assert_array_equal(result, expected_result)


def test_negative_all_negative_values():
    matrix = np.array([[-1, -2], [-3, -4]])
    expected_result = np.array([[-1, -2], [-3, -4]])
    result = _negative(matrix)
    np.testing.assert_array_equal(result, expected_result)


def test_negative_mixed_values():
    matrix = np.array([[1, -2], [-3, 4]])
    expected_result = np.array([[0, -2], [-3, 0]])
    result = _negative(matrix)
    np.testing.assert_array_equal(result, expected_result)


def test_negative_zero_values():
    matrix = np.array([[0, 0], [0, 0]])
    expected_result = np.array([[0, 0], [0, 0]])
    result = _negative(matrix)
    np.testing.assert_array_equal(result, expected_result)

# Tests for the _conf_mat_undirected function


def test_conf_mat_undirected_identical_graphs():
    truth = nx.Graph()
    truth.add_edges_from([("a", "b"), ("b", "c"), ("c", "a")])
    est = nx.Graph()
    est.add_edges_from([("a", "b"), ("b", "c"), ("c", "a")])
    feature_names = ["a", "b", "c"]
    Tp, Tn, Fp, Fn = _conf_mat_undirected(truth, est, feature_names)
    assert Tp == 3
    assert Tn == 0
    assert Fp == 0
    assert Fn == 0


def test_conf_mat_undirected_one_missing_edge():
    truth = nx.Graph()
    truth.add_edges_from([("a", "b"), ("b", "c"), ("c", "a")])
    est = nx.Graph()
    est.add_edges_from([("a", "b"), ("b", "c")])
    feature_names = ["a", "b", "c"]
    Tp, Tn, Fp, Fn = _conf_mat_undirected(truth, est, feature_names)
    assert Tp == 2
    assert Tn == 0
    assert Fp == 0
    assert Fn == 1


def test_conf_mat_undirected_one_extra_edge():
    truth = nx.Graph()
    truth.add_edges_from([("a", "b"), ("b", "c")])
    est = nx.Graph()
    est.add_edges_from([("a", "b"), ("b", "c"), ("c", "a")])
    feature_names = ["a", "b", "c"]
    Tp, Tn, Fp, Fn = _conf_mat_undirected(truth, est, feature_names)
    assert Tp == 2
    assert Tn == 0
    assert Fp == 1
    assert Fn == 0


def test_conf_mat_undirected_completely_different():
    truth = nx.Graph()
    truth.add_edges_from([("a", "b")])
    est = nx.Graph()
    est.add_edges_from([("b", "c")])
    feature_names = ["a", "b", "c"]
    Tp, Tn, Fp, Fn = _conf_mat_undirected(truth, est, feature_names)
    assert Tp == 0
    assert Tn == 1
    assert Fp == 1
    assert Fn == 1


def test_precision():
    # Test case where Tp = 0 and Fp = 0
    metrics = {'Tp': 0, 'Fp': 0}
    assert _precision(
        metrics) == 0, "Precision should be 0 when Tp and Fp are both 0"

    # Test case where Tp = 5 and Fp = 0
    metrics = {'Tp': 5, 'Fp': 0}
    assert _precision(
        metrics) == 1, "Precision should be 1 when Tp = 5 and Fp = 0"

    # Test case where Tp = 5 and Fp = 5
    metrics = {'Tp': 5, 'Fp': 5}
    assert _precision(
        metrics) == 0.5, "Precision should be 0.5 when Tp = 5 and Fp = 5"

    # Test case where Tp = 0 and Fp = 5
    metrics = {'Tp': 0, 'Fp': 5}
    assert _precision(
        metrics) == 0, "Precision should be 0 when Tp = 0 and Fp = 5"


def test_recall():
    # Test case where Tp = 0 and Fn = 0
    metrics = {'Tp': 0, 'Fn': 0}
    assert _recall(
        metrics) == 0, "Recall should be 0 when Tp and Fn are both 0"

    # Test case where Tp = 5 and Fn = 0
    metrics = {'Tp': 5, 'Fn': 0}
    assert _recall(metrics) == 1, "Recall should be 1 when Tp = 5 and Fn = 0"

    # Test case where Tp = 5 and Fn = 5
    metrics = {'Tp': 5, 'Fn': 5}
    assert _recall(
        metrics) == 0.5, "Recall should be 0.5 when Tp = 5 and Fn = 5"

    # Test case where Tp = 0 and Fn = 5
    metrics = {'Tp': 0, 'Fn': 5}
    assert _recall(metrics) == 0, "Recall should be 0 when Tp = 0 and Fn = 5"


def test_aupr():
    # Test case with perfect predictions
    metrics = {'_Gm': np.array([[1, 0], [0, 1]]),
               '_preds': np.array([[1, 0], [0, 1]])}
    assert _aupr(metrics) > 0.9, "AUPR should be high for perfect predictions"

    # Test case with random predictions
    metrics = {'_Gm': np.array([[1, 0], [0, 1]]),
               '_preds': np.array([[0.5, 0.5], [0.5, 0.5]])}
    assert 0 <= _aupr(
        metrics) <= 1, "AUPR should be between 0 and 1 for random predictions"

    # Test case with dimension mismatch
    metrics = {'_Gm': np.array([[1, 0]]), '_preds': np.array([[1, 0], [0, 1]])}
    assert _aupr(metrics) == 0.0, "AUPR should be 0 for dimension mismatch"


def test_f1():
    # Test case where precision = 0 and recall = 0
    metrics = {'precision': 0, 'recall': 0}
    assert _f1(metrics) == 0, "F1 should be 0 when precision and recall are both 0"

    # Test case where precision = 1 and recall = 1
    metrics = {'precision': 1, 'recall': 1}
    assert _f1(metrics) == 1, "F1 should be 1 when precision and recall are both 1"

    # Test case where precision = 0.5 and recall = 0.5
    metrics = {'precision': 0.5, 'recall': 0.5}
    assert _f1(
        metrics) == 0.5, "F1 should be 0.5 when precision and recall are both 0.5"



"""
1. **Imports**: Imported necessary modules and the `compute_score_mat_parallel`
function.
2. **Function Definition**: Translated the R function to Python, maintaining the same
logic and structure.
3. **Matrix Operations**: Used NumPy for matrix operations.
4. **Parallel Processing**: Used Python's `multiprocessing.Pool` for parallel
processing, similar to R's `mcmapply`.

Make sure to have the `compute_score_mat_parallel` function defined in a file named
`compute_score_mat_parallel.py`.
"""

# pylint: disable=E1101:no-member, W0201:attribute-defined-outside-init, W0511:fixme
# pylint: disable=C0103:invalid-name, W0221:arguments-differ
# pylint: disable=C0116:missing-function-docstring
# pylint: disable=R0913:too-many-arguments, E0401:import-error
# pylint: disable=R0914:too-many-locals, R0915:too-many-statements
# pylint: disable=W0106:expression-not-assigned, R1702:too-many-branches

from multiprocessing import Pool

import numpy as np

from .computeScoreMatParallel import computeScoreMatParallel


def updateScoreMat(
        score_mat,
        X,
        score_name,
        i,
        j,
        score_nodes,
        adj,
        verbose,
        num_cores,
        max_num_parents,
        pars_score,
        interv_mat,
        interv_data):
    """_summary_

    Args:
        score_mat (_type_): _description_
        X (_type_): _description_
        score_name (_type_): _description_
        i (_type_): _description_
        j (_type_): _description_
        score_nodes (_type_): _description_
        adj (_type_): _description_
        output (_type_): _description_
        num_cores (_type_): _description_
        max_num_parents (_type_): _description_
        pars_score (_type_): _description_
        interv_mat (_type_): _description_
        interv_data (_type_): _description_

    Returns:
        _type_: _description_
    """
    # new edge: from i to j

    p = X.shape[1]
    if verbose:
        print(f"p: {p}")
        print("Adjacency matrix---")
        for r in range(adj.shape[0]):
            for c in range(p):
                print(f"{adj[r, c]} ", end="")
            print()

    existing_par_of_j = np.where(adj[:, j] == 1)[0]
    not_allowed_par_of_j = np.setdiff1d(
        np.where(score_mat[:, j] == -np.inf)[0], np.append(existing_par_of_j, j))
    if verbose:
        print(f"existing_par_of_j: {existing_par_of_j}")
        print(f"not_allowed_par_of_j: {not_allowed_par_of_j}")

    # if there is something left that we need to update
    if len(existing_par_of_j) + len(not_allowed_par_of_j) < p - 1:
        # update column for j
        row_parents = np.full((p, len(existing_par_of_j) + 1), np.nan)
        row_parents[:, :len(existing_par_of_j)] = existing_par_of_j
        row_parents[:, len(existing_par_of_j)] = np.arange(p)
        to_update = np.setdiff1d(np.arange(p), np.append(
            [j], np.append(existing_par_of_j, not_allowed_par_of_j)))

        if len(existing_par_of_j) < max_num_parents:
            if num_cores == 1:
                score_update = [
                    computeScoreMatParallel(
                        row_parents=row_parents,
                        sel_mat=np.ones((p, p), dtype=bool),
                        score_name=score_name, X=X, verbose=verbose, node2=j,
                        pars_score=pars_score, interv_mat=interv_mat,
                        interv_data=interv_data, i=idx) \
                            for idx in to_update
                    ]
            else:
                with Pool(num_cores) as pool:
                    score_update = pool.starmap(
                        computeScoreMatParallel, [
                            (row_parents, np.ones((p, p), dtype=bool),
                             score_name, X, verbose, j, pars_score, interv_mat,
                             interv_data, idx) \
                                 for idx in to_update
                        ])
        else:
            score_update = -np.inf

        score_mat[to_update, j] = score_update - score_nodes[j]

    return score_mat
